from __future__ import annotations
from typing import Iterable
import numpy as np
from sfHMM.motor.base import sfHMMmotorBase
from sfHMM.single_sfhmm import sfHMM1, _S
from sfHMM.multi_sfhmm import sfHMMn
from sfHMM.utils import append_log, concat, sfHMMAnalysisError


class sfHMM1Motor(sfHMMmotorBase, sfHMM1):
    """sfHMM for motor stepping.

    Parameters
    ----------
    data_raw : array like, optional
        Data for analysis.
    sg0 : float, optional
        Parameter used in filtering method. Expected to be 20% of signal change.
        If <= 0, sg0 will be determined automatically.
    psf : float, optional
        Transition probability used in step finding algorithm.
        if 0 < p < 0.5 is not satisfied, the original Kalafut-Visscher's algorithm is
        executed.
    krange : int or (int, int)
        Minimum and maximum number of states to search in GMM clustering. If it is
        integer, then it will be interpretted as (krange, krange).
    model: str, default "g" (= Gaussian)
        Distribution of noise. Gauss and Poisson distribution are available for now.
    name : str, optional
        Name of the object.
    max_strides : int, default 2
        The largest step of motor. If max_stride = 2, then from 2-step backward to
        2-step forward steps are considered. Larger value results in longer calculation
        time.
    """

    def __init__(
        self,
        data_raw: _S = None,
        *,
        sg0: float = -1,
        psf: float = -1,
        krange: int | tuple[int, int] | None = None,
        model: str = "g",
        name: str = "",
        max_stride: int = 2,
    ):
        super().__init__(sg0=sg0, psf=psf, krange=krange, model=model, name=name)
        self.data_raw = data_raw
        self.max_stride = max_stride
        self.covariance_type = "tied"

    def _set_covars(self):
        if self.states is None:
            raise sfHMMAnalysisError(
                "Cannot initialize 'covars_' because the state sequence "
                "'states' has yet been determined."
            )
        self.covars_ = [[np.var(self.data_raw - self.step.fit)]]
        self.min_covar = np.min(self.covars_) * 0.015
        return None

    def _set_transmat(self):
        if self.states is None:
            raise sfHMMAnalysisError(
                "Cannot initialize 'transmat_' because the state sequence "
                "'states' hasyet been determined."
            )
        transmat_kernel = np.zeros(self.max_stride * 2 + 1)
        dy = np.diff(self.states)
        dy = dy[np.abs(dy) <= self.max_stride]
        dy_unique, counts = np.unique(dy, return_counts=True)
        transmat_kernel[dy_unique + self.max_stride] = counts
        transmat_kernel += 1e-12
        self.transmat_kernel = transmat_kernel / np.sum(transmat_kernel)

        return None

    def _estimate_krange(self, estimation: str):
        dy = self._accumulate_step_sizes()
        nsmall, nlarge = sorted(map(int, [np.sum(dy > 0), np.sum(dy < 0)]))
        if estimation == "fast":
            self.krange = (nlarge - nsmall, nlarge - nsmall // 2)
        elif estimation == "safe":
            self.krange = (nlarge - nsmall, nlarge)
        else:
            raise ValueError(f"Cannot interpret estimation method: {estimation}")
        return None


class sfHMMnMotor(sfHMMmotorBase, sfHMMn):
    """sfHMM for multi-trajectory motor stepping.

    Parameters
    ----------
    data_raw : iterable, optional
        Datasets to be added.
    sg0 : float, optional
        Parameter used in filtering method. Expected to be 20% of signal change.
        If <= 0, sg0 will be determined automatically.
    psf : float, optional
        Transition probability used in step finding algorithm.
        if 0 < p < 0.5 is not satisfied, the original Kalafut-Visscher's algorithm is executed.
    krange : int or (int, int)
        Minimum and maximum number of states to search in GMM clustering. If it is integer, then
        it will be interpretted as (krange, krange).
    model: str, by default "g" (= Gaussian)
        Distribution of noise. Gauss and Poisson distribution are available for now.
    name : str, optional
        Name of the object.
    max_strides : int, default is 2.
        The largest step of motor. If max_stride = 2, then from 2-step backward to 2-step
        forward steps are considered. Larger value results in longer calculation time.
    """

    def __init__(
        self,
        data_raw: Iterable[_S] = None,
        *,
        sg0: float = -1,
        psf: float = -1,
        krange: int | tuple[int, int] | None = None,
        model: str = "g",
        name: str = "",
        max_stride: int = 2,
    ):
        super().__init__(sg0=sg0, psf=psf, krange=krange, model=model, name=name)
        self.max_stride = max_stride
        self.covariance_type = "tied"
        if data_raw is not None:
            self.appendn(data_raw)

    @append_log
    def append(self, data: _S, name: str = None) -> sfHMMnMotor:
        if name is None:
            name = self.name + f"[{self.n_data}]"

        sf = sfHMM1Motor(
            data,
            sg0=self.sg0,
            psf=self.psf,
            krange=self.krange,
            model=self.model,
            name=name,
            max_stride=self.max_stride,
        )
        self.n_data += 1
        self._sf_list.append(sf)
        self.ylim[0] = min(sf.ylim[0], self.ylim[0])
        self.ylim[1] = max(sf.ylim[1], self.ylim[1])
        return self

    def _set_covars(self):
        step_fit = np.array(concat([sf.step.fit for sf in self]))
        self.covars_ = [[np.var(self.data_raw - step_fit)]]
        return None

    def _set_transmat(self):
        transmat_kernel = np.zeros(self.max_stride * 2 + 1)
        dy = np.array(concat([np.diff(sf.states) for sf in self]))
        dy = dy[np.abs(dy) <= self.max_stride]
        dy_unique, counts = np.unique(dy, return_counts=True)
        transmat_kernel[dy_unique + self.max_stride] = counts
        transmat_kernel += 1e-12
        self.transmat_kernel = transmat_kernel / np.sum(transmat_kernel)

        return None

    def _copy_params(self, sf: sfHMM1Motor):
        if self.covariance_type == "spherical":
            sf.covars_ = self.covars_.ravel()
        else:
            sf.covars_ = [[self.covars_[0, 0, 0]]]
        sf.min_covar = self.min_covar
        sf.means_ = self.means_
        sf.startprob_ = self.startprob_
        sf.transmat_kernel = self.transmat_kernel
        return None

    def _estimate_krange(self, estimation: str):
        nsmall = nlarge = 0
        for sf in self:
            dy = sf._accumulate_step_sizes()
            nsmall_, nlarge_ = sorted(map(int, [np.sum(dy > 0), np.sum(dy < 0)]))
            nsmall = max(nsmall, nsmall_)
            nlarge = max(nlarge, nlarge_)

        if estimation == "fast":
            self.krange = (nlarge - nsmall, nlarge - nsmall // 2)
        elif estimation == "safe":
            self.krange = (nlarge - nsmall, nlarge)
        else:
            raise ValueError(f"Cannot interpret estimation method: {estimation}")
        return None



__version__ = "0.7.0"

from sfHMM.single_sfhmm import sfHMM1
from sfHMM.multi_sfhmm import sfHMMn
from sfHMM import io
from sfHMM.sampling import hmm_sampling, motor_sampling
from sfHMM import motor


# Inheritance Map
# ---------------

#       (hmmlearn.hmm.GaussianHMM)
#                   |
#                   |
#              (sfHMMBase)
#            /      |      \
#          /        |        \
#    sfHMM1  (sfHMMmotorBase)  sfHMMn
#        \        /   \        /
#         \     /       \     /
#      sfHMM1Motor    sfHMMnMotor

__all__ = ["sfHMM1", "sfHMMn", "io", "hmm_sampling", "motor_sampling", "motor"]



import numpy as np
from dataclasses import dataclass
from scipy.special import gammaln, logsumexp

# Ideally we should prepare `n = np.arange(1, len(data))` first, and view
# it many times in get_optimal_splitter like n[:len(self.fw)], but this
# did not improve efficiency so much.


@dataclass
class Moment:
    """
    This class aims at splitting arrays in a way that moment can be calculated very
    fast.

    fw          bw
     0 o ++++++ 0
     1 oo +++++ 1
     : :      : :
     n oooooo + n

    When ndarray `data` is given by `self.init(data)`, self.fw[i,k] means
    np.sum(data[:k]**(i+1)), while self.bw[i,k] means np.sum(data[-k:]**(i+1)).
    self.total[i] means np.sum(data**(i+1)).
    """

    fw: np.ndarray = None
    bw: np.ndarray = None
    total: np.ndarray = None

    def __len__(self):
        """
        Get the length of the constant region
        """
        return self.fw.shape[1] + 1

    def complement(self):
        """
        Complement moment values.
        Calculate fw from bw, or vice versa.


        fw      bw=None
         0 o
         1 oo
         : :
         k oooo
                    +++++
                     ++++
                        :
                        +
              fw=None   bw

        becomes

        fw      bw
         0 o ++++
         1 oo +++
         : :
         k oooo +
                  o +++++
                  oo ++++
                        :
                  ooooo +
                 fw     bw

        """
        if self.fw is None:
            self.fw = self.total.reshape(-1, 1) - self.bw
        elif self.bw is None:
            self.bw = self.total.reshape(-1, 1) - self.fw
        else:
            pass
        return None

    def split(self, i: int):
        """
        Split a Moment object into to Moment objects at position i.
        This means :i-1 will be the former, and i: will be the latter.

        fw          bw
         0 o ++++++++++++ 0
         1 oo +++++++++++ 1
         : :            : :
           ooooooooooo ++
         n oooooooooooo + n

        becomes

        fw      bw=None
         0 o
         1 oo
         : :
         k oooooo
                  +++++++
                   ++++++
                        :
                        +
              fw=None   bw
        """
        m1 = self.__class__(self.fw[:, : i - 1], None, self.fw[:, i - 1])
        m2 = self.__class__(None, self.bw[:, i:], self.bw[:, i - 1])
        m1.complement()
        m2.complement()
        return (m1, m2)

    def init(self, data, order=1):
        """
        Initialization using total data.

        Parameters
        ----------
        data : array
            The input data.
        order : int, optional
            Up to what order of moment should be calculated, by default 1
        """
        self.__class__.fwn = np.arange(1, len(data))
        self.__class__.bwn = np.arange(len(data) - 1, 0, -1)
        orders = np.arange(1, order + 1)
        self.fw = np.vstack([np.cumsum(data[:-1] ** o) for o in orders])
        self.total = np.array(self.fw[:, -1] + data[-1] ** orders)
        self.complement()
        return self

    def get_optimal_splitter(self):
        pass


@dataclass
class GaussMoment(Moment):
    @property
    def chi2(self):
        return self.total[1] - self.total[0] ** 2 / len(self)

    def get_optimal_splitter(self):
        n = np.arange(1, len(self))
        chi2_fw = self.fw[1] - self.fw[0] ** 2 / n
        chi2_bw = self.bw[1] - self.bw[0] ** 2 / n[::-1]
        chi2 = chi2_fw + chi2_bw
        x = np.argmin(chi2)
        return chi2[x] - self.chi2, x + 1

    def init(self, data):
        return super().init(data, order=2)


@dataclass
class SDFixedGaussMoment(Moment):
    @property
    def sq(self):
        return (2 - 1 / len(self)) / len(self) * self.total[0] ** 2

    def get_optimal_splitter(self):
        n = np.arange(1, len(self))
        sq_fw = (2 - 1 / n) / n * self.fw[0] ** 2
        sq_bw = (2 - 1 / n[::-1]) / n[::-1] * self.bw[0] ** 2
        sq = sq_fw + sq_bw
        x = np.argmax(sq)
        return sq[x] - self.sq, x + 1


@dataclass
class TtestMoment(Moment):
    def get_optimal_splitter(self):
        n = np.arange(1, len(self))
        tk = np.abs(self.fw[0] / n - self.bw[0] / n[::-1]) / np.sqrt(
            1 / n + 1 / (n[::-1])
        )
        x = np.argmax(tk)
        return tk[x], x + 1


@dataclass
class PoissonMoment(Moment):
    @property
    def slogm(self):
        return self.total[0] * np.log((self.total[0] + 1e-12) / len(self))

    def get_optimal_splitter(self):
        n = np.arange(1, len(self))
        slogm_fw = self.fw[0] * np.log((self.fw[0] + 1e-12) / n)
        slogm_bw = self.bw[0] * np.log((self.bw[0] + 1e-12) / n[::-1])
        slogm = slogm_fw + slogm_bw
        x = np.argmax(slogm)
        return slogm[x] - self.slogm, x + 1


@dataclass
class BayesianPoissonMoment(Moment):
    def get_optimal_splitter(self):
        n = np.arange(1, len(self))
        g1 = gammaln(self.fw[0] + 1)
        g2 = gammaln(self.bw[0] + 1)
        logprob = (
            g1
            + g2
            - (self.fw[0] + 1) * np.log(n)
            - (self.bw[0] + 1) * np.log(n[::-1])
            - np.log((self.fw[0] / n) ** 2 + (self.bw[0] / n[::-1]) ** 2)
        )
        logC = (
            np.log(2 / np.pi / (len(self) - 1))
            - gammaln(self.total[0])
            + self.total[0] * np.log(len(self))
        )
        logBayesFactor = logC + logsumexp(logprob)
        return logBayesFactor, np.argmax(logprob) + 1



from __future__ import annotations

from typing import TYPE_CHECKING
from pathlib import Path
import numpy as np
from sfHMM.io.utils import check_ref, infer_sep
from sfHMM.base import sfHMMBase
from sfHMM.single_sfhmm import sfHMM1
from sfHMM.multi_sfhmm import sfHMMn
from warnings import warn

if TYPE_CHECKING:
    import pandas as pd

__all__ = ["read", "read_excel", "save"]

EXCEL_TESTED = (".xlsx", ".xlsm", ".xltx", ".xltm", ".xls")
_EXCEL_SUFFIXES = frozenset(
    [
        ".xlsx",
        ".xlsm",
        ".xlsb",
        ".xltx",
        ".xltm",
        ".xls",
        ".xlt",
        ".xml",
        ".xlam",
        ".xla",
        ".xlw",
        ".xlr",
    ]
)


def read(
    path: str,
    out: sfHMMn | None = None,
    sep: str = None,
    encoding: str = None,
    header: int = 0,
    **kwargs,
) -> sfHMMn:
    """
    Read a file using appropriate function, and import its data to sfHMMn object. This
    function can read many types of files.

    Parameters
    ----------
    path : str
        Path to file.
    out : sfHMMn or its subclass object, optional
        In which object the content of the file will be stored. If None, a new sfHMMn object
        with default setting will be made. This argument is useful when you prefer your
        own setting of sfHMM input parameter(s).
    sep, encoding, header
        Important arguments. Default is header=0 rather than "infer" because header="infer"
        usually works in a wrong way.
    **kwargs
        Other keyword arguments that will passed to pd.read_csv() or pd.read_excel().

    Returns
    -------
    sfHMMn object
        Object with datasets.
    """
    import pandas as pd

    out = check_ref(out, sfHMMn)
    if is_excel_file(path):
        df = _safe_read_excel(path, sheet_name=0, header=header, **kwargs)
    else:
        # infer sep if it was not given
        sep = infer_sep(path, encoding) if sep is None else sep
        df = pd.read_csv(path, sep=sep, encoding=encoding, header=header, **kwargs)
    out.from_pandas(df)
    return out


def read_excel(
    path: str,
    ref: sfHMMBase = None,
    ignore_exceptions: bool = True,
    header: int = 0,
    squeeze: bool = False,
    **kwargs,
) -> dict[str, sfHMMn] | sfHMMn:
    """
    Read a Excel file using pandas.read_excel, and import its data to sfHMMn object. Every sheet
    must have only single trajectory.

    Parameters
    ----------
    path : str
        Path to Excel file.
    ref : sfHMMBase or its subclass object, optional
        From which object the input parameters will be referred. If None, new sfHMMn objects
        with default setting will be made every time. This argument is useful when you prefer
        your own setting of sfHMM input parameter(s).
    ignore_exceptions : bool
        If True, exceptions will be ignored and only sheets with valid format will be included
        in the output list.
    header : int, default is 0
        Header index.
    squeeze: bool, default is False
        If Excel file only contains one sheet, then return sfHMMn object instead of redundant
        dictionary like {"Sheet1": <sfHMMn>}. Default is set to False for compatibility.
        Although this argument collides with that in pd.read_excel(), it is not a problem
        because squeeze in pd.read_excel must be False in our usage.
    **kwargs
        Other keyword arguments that will passed to pd.read_excel().

    Returns
    -------
    sfHMMn objects
        Dictionary of objects with datasets
    """
    ext = Path(path).suffix
    if ext in (".csv", ".txt", ".dat"):
        raise ValueError("For csv, txt or dat files, use `read()` function instead.")

    ref = check_ref(ref, sfHMMBase)

    df_dict = _safe_read_excel(path, sheet_name=None, header=header, **kwargs)

    msfdict = {}
    for sheet_name, df in df_dict.items():
        msf = sfHMMn(**ref.get_params(), name=sheet_name)
        try:
            msf.from_pandas(df)
            msfdict[sheet_name] = msf
        except Exception:
            if ignore_exceptions:
                pass
            else:
                raise

    if len(msfdict) == 0:
        raise RuntimeError("No sfHMMn object was successfully made.")
    elif squeeze and len(msfdict) == 1:
        msfdict = msfdict[sheet_name]

    if ext not in EXCEL_TESTED:
        warn(
            f"Extension '{ext}' has yet been tested. "
            f"{', '.join(EXCEL_TESTED)} are recommended.",
            UserWarning,
        )

    return msfdict


def save(obj: sfHMMBase, path: str) -> None:
    """
    Save obj.step.fit, obj.data_fil and obj.viterbi as csv.

    Parameters
    ----------
    obj : sfHMMBase
        sfHMM object to save.
    path : str
        Saving path.
    """
    import pandas as pd

    df_list = _to_dataframes(obj)
    out = pd.concat(df_list, axis=1)
    out.to_csv(path)
    return None


def is_excel_file(path: str | Path):
    return Path(path).suffix in _EXCEL_SUFFIXES


def _safe_read_excel(path, **kwargs):
    # xlrd cannot open xlsx in some versions. Here try openpyxl if possible.
    import pandas as pd

    try:
        out = pd.read_excel(path, **kwargs)
    except Exception:
        if kwargs.get("engine", None) == "openpyxl":
            raise
        kwargs["engine"] = "openpyxl"
        out = pd.read_excel(path, **kwargs)
    return out


def _to_dataframes(obj: sfHMMBase, suffix: str = "") -> list[pd.DataFrame]:
    import pandas as pd

    if isinstance(obj, sfHMM1):
        df = pd.DataFrame(
            data=obj.data_raw,
            dtype=np.float64,
            columns=[f"data_raw{suffix}"],
            index=np.arange(obj.data_raw.size, dtype=np.int32),
        )
        if obj.step is not None:
            df[f"step finding{suffix}"] = obj.step.fit
        if obj.data_fil is not None:
            df[f"denoised{suffix}"] = obj.data_fil
        if obj.viterbi is not None:
            df[f"Viterbi path{suffix}"] = obj.viterbi
        return [df]

    elif isinstance(obj, sfHMMn):
        df_list = []
        for i, sf in enumerate(obj):
            df_list.append(_to_dataframes(sf, suffix=f"-{i}")[0])
        return df_list
    else:
        raise TypeError(
            f"Only sfHMM objects can be converted to pd.DataFrame, but got {type(obj)}"
        )



from __future__ import annotations
from typing import Iterable, Union
import numpy as np
import pandas as pd
import matplotlib as mpl
from matplotlib.axes import Axes
import matplotlib.pyplot as plt
from qtpy.QtWidgets import (
    QGridLayout,
    QWidget,
    QCheckBox,
    QMainWindow,
    QDockWidget,
    QSpinBox,
    QAction,
    QApplication,
)
from qtpy.QtCore import Qt
from .canvas import EventedCanvas

DictLike = Union[dict, pd.DataFrame]

APPLICATION = None


def type_check(data) -> dict[str, np.ndarray]:
    if isinstance(data, pd.DataFrame):
        data = data.to_dict()
    elif isinstance(data, dict):
        data = {k: np.asarray(v) for k, v in data.items()}
    else:
        raise TypeError(f"data must be dict or pd.DataFrame, got {type(data)}")
    return data


class TrajectoryViewer(QMainWindow):
    """
    Show interactive multi-trajectory plot.

    >>> data = {"traj-1": np.array([...]),
                "traj-2": np.array([...])
                }
    >>> viewer = TrajectoryViewer(data)
    >>> viewer.show()
    """

    def __init__(
        self,
        data: DictLike | list[DictLike] | tuple[DictLike],
        styles: dict = None,
        colors: dict = None,
    ):
        app = get_app()  # noqa: F841
        super().__init__(parent=None)
        # check input
        if isinstance(data, (tuple, list)):
            data = list(map(type_check, data))
        else:
            data = [type_check(data)]

        self.data = data
        self.checkbox = Controller(self, data[0].keys())

        self.addDockWidget(Qt.BottomDockWidgetArea, self.checkbox)
        self.checkbox.visibilityChanged.connect(self.checkboxes_visibility_changed)

        self.setUnifiedTitleAndToolBarOnMac(True)

        self.plot_style = styles.copy() if styles is not None else {}
        self.plot_style["legend.frameon"] = True
        self.plot_color = colors.copy() if colors is not None else {}
        self.lines: dict[str, plt.Line2D] = dict()
        self.current_index = 0
        # prepare a figure canvas
        # To block additional figure being opened, matplotlib backend must be changed temporary.
        backend = mpl.get_backend()
        mpl.use("Agg")
        with plt.style.context(self.plot_style):
            self.fig = plt.figure()
            self.ax: Axes = self.fig.add_subplot(111)
            for key, data in self.current_data.items():
                self.lines[key] = self.ax.plot(
                    data, color=self.plot_color[key], label=key
                )[0]

        mpl.use(backend)

        self.setCentralWidget(EventedCanvas(self.fig))
        self.fig.tight_layout()
        self.fig.canvas.draw()

        self.menu = self.menuBar().addMenu("&Menu")

        show_legend = QAction("Show legend", parent=self, checkable=True, checked=False)
        show_legend.triggered.connect(self.change_legend_visibility)
        self.menu.addAction(show_legend)

        self.show_controller = QAction(
            "Show checkboxes", parent=self, checkable=True, checked=True
        )
        self.show_controller.triggered.connect(self.change_checkboxes_visibility)
        self.menu.addAction(self.show_controller)

        close = QAction("Close", parent=self)
        close.triggered.connect(self.close)
        self.menu.addAction(close)

    @property
    def current_data(self) -> dict[str, np.ndarray]:
        """
        Return the dataset dict that should be displayed. Independent of whether they are
        visible or not.
        """
        return self.data[self.current_index]

    def update_plot(self):
        checked = self.checkbox.isEachChecked()
        for i, key in enumerate(self.current_data.keys()):
            if checked[i]:
                self.lines[key].set_color(self.plot_color[key])
            else:
                self.lines[key].set_color([0, 0, 0, 0])

        self.fig.tight_layout()
        self.fig.canvas.draw()

    def change_data(self):
        for key, value in self.current_data.items():
            self.lines[key].set_xdata(np.arange(len(value)))
            self.lines[key].set_ydata(value)

        self.fig.tight_layout()
        self.fig.canvas.draw()

    def change_checkboxes_visibility(self):
        if self.show_controller.isChecked():
            self.checkbox.show()
        else:
            self.checkbox.hide()

    def checkboxes_visibility_changed(self):
        self.show_controller.setChecked(self.checkbox.isVisible())
        return None

    def change_legend_visibility(self):
        legend = self.ax.get_legend()
        if legend:
            legend.remove()
        else:
            with plt.style.context(self.plot_style):
                # temporary set all the lines visible.
                # Without this code, lines will not be displayed in legend.
                for key in self.current_data.keys():
                    self.lines[key].set_color(self.plot_color[key])
                self.ax.legend()

        self.update_plot()

    def show(self):
        super().show()
        self.raise_()
        self.activateWindow()
        return None

    def closeEvent(self, event):
        """
        When QMainWindow is closing, its dock widget will not automatically close when it is
        floating. We have to catch the close event and disable floating before actually close
        the main window.
        """
        for dock in self.findChildren(Controller):
            dock.setFloating(False)

        event.accept()


class Controller(QDockWidget):
    def __init__(self, parent: TrajectoryViewer, names: Iterable[str]):
        super().__init__("Trajectories", parent=parent)

        self.setAllowedAreas(Qt.BottomDockWidgetArea | Qt.RightDockWidgetArea)
        self.qcheckbox_list: list[QCheckBox] = []

        layout = QGridLayout()
        layout.setAlignment(Qt.AlignCenter)
        self.central_widget = QWidget(self)
        self.central_widget.setLayout(layout)

        self._set_checkboxes(names)
        self._set_spinbox()
        self.setWidget(self.central_widget)

    def parent(self) -> TrajectoryViewer:  # for IDE
        return super().parent()

    def isEachChecked(self) -> list[bool]:
        return [widget.isChecked() for widget in self.qcheckbox_list]

    def _set_checkboxes(self, names):
        for name in names:
            checkbox = QCheckBox(self)
            checkbox.setText(name)
            checkbox.setChecked(True)
            checkbox.stateChanged.connect(self.parent().update_plot)
            self.qcheckbox_list.append(checkbox)
            self.central_widget.layout().addWidget(checkbox)

    def _set_spinbox(self):
        if len(self.parent().data) == 1:
            return None
        self.spinbox = QSpinBox(parent=self)
        self.spinbox.setRange(0, len(self.parent().data) - 1)

        @self.spinbox.valueChanged.connect
        def _():
            self.parent().current_index = int(self.spinbox.value())
            self.parent().change_data()

        self.central_widget.layout().addWidget(self.spinbox)


def gui_qt():
    try:
        from IPython import get_ipython
    except ImportError:
        return None

    shell = get_ipython()

    if shell and shell.active_eventloop != "qt":
        shell.enable_gui("qt")
    return None


def get_app():
    gui_qt()
    app = QApplication.instance()
    if app is None:
        app = QApplication([])
    global APPLICATION
    APPLICATION = app
    return app



import numpy as np
import matplotlib.pyplot as plt
from sfHMM.step.moment import (
    Moment,
    GaussMoment,
    SDFixedGaussMoment,
    TtestMoment,
    PoissonMoment,
    BayesianPoissonMoment,
)
from scipy.stats import t as student_t
from scipy.stats import norm
import heapq

r"""
    An efficient, extended version of step finding algorithm original in [1].
    Although similar summation of data is repeated many times during a single run, the
original algorithm did not make full use of the results obtained in previous iterations.
In short, before calculating such as square deviation, the 'generator' of it is stored and
passed to next iteration. The n-th-order moment E[X_i^n] is very useful for in this purpose,
that's why `Moment` class is defined in moment.py. A `Moment` object corresponds to one constant
region (between two signal change points) and has n-th order moment with any step position as
attributes. Here in step finding classes, a corresponding `Moment` object is split into two
fragments and they are passed to newly generated two steps.
    In this optimized implementation, both GaussStep and PoissonStep can analyze 100,000-point
data within 1 sec!

Notes
-----
`x0` and `dx` always denote positions shown below:

       x0+dx
 0  x0   v  oooooooooo
 v   v   ooo    <- new step (2)
 oooo
     .......    <- old step
     oooo       <- new step (1)

Inheritance Map
---------------
                         BaseStep
                      /            \
           RecursiveStep            GaussStep
            /       \
SDFixedGaussStep, PoissonStep,
TtestStep, BayesianPoissonStep

"""


class Heap:
    """
    Priorty queue. I wrapped heapq because it is not intuitive.
    """

    def __init__(self):
        self.heap = []

    def push(self, item):
        heapq.heappush(self.heap, item)

    def pop(self):
        return heapq.heappop(self.heap)


def estimate_sigma(data):
    p = norm.cdf(1)  # = sigma for standard normal distribution.
    return np.quantile(np.diff(data), p) / np.sqrt(2)


class BaseStep:
    def __init__(self, data, p):
        self.data = np.asarray(data)
        self.len = self.data.size
        self.n_step = 1
        self.step_list = [0, self.len]
        if 0.0 < p < 0.5:
            self.p = p
        else:
            self.p = 1 / (1 + np.sqrt(self.len))

        self.penalty = np.log(self.p / (1 - self.p))

    def multi_step_finding(self):
        """
        Run step-finding algorithm and store all the information.
        """

    def _finalize(self):
        """
        Called at the end of fitting.
        """
        self.step_list.sort()
        self.n_step = len(self.step_list) - 1
        self.mu_list = np.zeros(self.n_step)

        for i in range(self.n_step):
            self.mu_list[i] = np.mean(
                self.data[self.step_list[i] : self.step_list[i + 1]]
            )
            self.fit[self.step_list[i] : self.step_list[i + 1]] = self.mu_list[i]

        self.len_list = np.diff(self.step_list)
        self.step_size_list = np.diff(self.mu_list)

        return None

    def plot(self):
        plt.plot(self.data, color="lightgray", label="raw data")
        plt.plot(self.fit, color="red", label="fit")
        plt.legend(bbox_to_anchor=(1.05, 1), loc="upper left", borderaxespad=0)
        plt.show()
        return None


class RecursiveStep(BaseStep):
    def _append_steps(self, mom, x0: int = 0):
        if len(mom) < 3:
            return None
        s, dx = mom.get_optimal_splitter()
        if self._continue(s):
            self.step_list.append(x0 + dx)
            mom1, mom2 = mom.split(dx)
            self._append_steps(mom1, x0=x0)
            self._append_steps(mom2, x0=x0 + dx)
        else:
            pass
        return None

    def _init_momemt(self) -> Moment: ...

    def _continue(self, s) -> bool: ...

    def multi_step_finding(self):
        mom = self._init_moment()
        self.fit = np.full(self.len, mom.total[0] / self.len)
        self._append_steps(mom)
        self._finalize()
        return self


class GaussStep(BaseStep):
    """
    Gauss-distribution step finding.

    Reference
    ---------
    Kalafut, B., & Visscher, K. (2008). An objective, model-independent method for
    detection of non-uniform steps in noisy signals. Computer Physics Communications,
    179(10), 716–723. https://doi.org/10.1016/j.cpc.2008.06.008
    """

    def __init__(self, data, p=-1):
        """
        Parameters
        ----------
        data : array
            Input array.
        p : float, optional
            Probability of transition (signal change). If not in a proper range 0<p<0.5,
            then This algorithm will be identical to the original Kalafut-Visscher's.
        """
        super().__init__(np.asarray(data, dtype=np.float64), p)

    def multi_step_finding(self):
        g = GaussMoment().init(self.data)
        self.fit = np.full(self.len, g.total[0] / self.len)
        chi2 = g.chi2  # initialize total chi^2
        heap = Heap()  # chi^2 change (<0), dx, x0, GaussMoment object of the step
        heap.push(g.get_optimal_splitter() + (0, g))

        while True:
            dchi2, dx, x0, g = heap.pop()
            dlogL = self.penalty - self.len / 2 * np.log(1 + dchi2 / chi2)

            if dlogL > 0:
                x = x0 + dx
                g1, g2 = g.split(dx)
                len(g1) > 2 and heap.push(g1.get_optimal_splitter() + (x0, g1))
                len(g2) > 2 and heap.push(g2.get_optimal_splitter() + (x, g2))
                self.step_list.append(x)
                chi2 += dchi2
            else:
                break

        self._finalize()
        return self


class SDFixedGaussStep(RecursiveStep):
    """
    Gauss-distribution step finding with fixed standard deviation of noise.
    If standard deviation of noise is unknown then it will be estimated by
    wavelet method. Compared to GaussStep, this algorithm detects more steps
    in some cases and less in others.

    """

    def __init__(self, data, p=-1, sigma=-1):
        super().__init__(np.asarray(data, dtype=np.float64), p)
        if sigma < 0:
            sigma = estimate_sigma(data)
        self.sigma = sigma

    def _init_moment(self):
        return SDFixedGaussMoment().init(self.data)

    def _continue(self, sq) -> bool:
        return self.penalty + sq / (2 * self.sigma**2) > 0


class TtestStep(RecursiveStep):
    """
    T-test based step finding.

    Reference
    ---------
    Shuang, B., Cooper, D., Taylor, J. N., Kisley, L., Chen, J., Wang, W., ... & Landes,
    C. F. (2014). Fast step transition and state identification (STaSI) for discrete
    single-molecule data analysis. The journal of physical chemistry letters, 5(18), 3157-3161.
    https://doi.org/10.1021/jz501435p
    """

    def __init__(self, data, alpha=0.05, sigma=-1):
        self.data = np.asarray(data, dtype=np.float64)
        self.len = self.data.size
        self.n_step = 1
        self.step_list = [0, self.len]
        if not 0 < alpha < 0.5:
            alpha = 0.05
        self.alpha = alpha

        if sigma < 0:
            sigma = estimate_sigma(data)

        self.sigma = sigma

    def _init_moment(self):
        return TtestMoment().init(self.data)

    def _append_steps(self, mom, x0: int = 0):
        if len(mom) < 3:
            return None
        tk, dx = mom.get_optimal_splitter()
        t_cri = student_t.ppf(1 - self.alpha / 2, len(mom))
        if t_cri < tk / self.sigma:
            self.step_list.append(x0 + dx)
            mom1, mom2 = mom.split(dx)
            self._append_steps(mom1, x0=x0)
            self._append_steps(mom2, x0=x0 + dx)
        else:
            pass
        return None


class PoissonStep(RecursiveStep):
    """
    Poisson distribution step finding. Input must be integer.
    """

    def __init__(self, data, p=-1):
        super().__init__(data, p)
        if not np.issubdtype(self.data.dtype, np.integer):
            raise TypeError("In PoissonStep, non-integer data type is forbidden.")
        elif np.any(self.data < 0):
            raise ValueError("Input data contains negative values.")

    def _init_moment(self):
        return PoissonMoment().init(self.data)

    def _continue(self, dlogL):
        return self.penalty + dlogL > 0


class BayesianPoissonStep(RecursiveStep):
    """
    Poisson distribution step finding in a Bayesian method.

    Reference
    ---------
    Ensign, D. L., & Pande, V. S. (2010). Bayesian detection of intensity changes in
    single molecule and molecular dynamics trajectories. Journal of Physical Chemistry
    B, 114(1), 280–292. https://doi.org/10.1021/jp906786b
    """

    def __init__(self, data, skept=4):
        self.data = np.asarray(data)
        self.len = self.data.size
        self.n_step = 1
        self.step_list = [0, self.len]
        if skept <= 0:
            raise ValueError(f"`skept` must be larger than 0, but got {skept}")
        self.skept = skept
        if not np.issubdtype(self.data.dtype, np.integer):
            raise TypeError("In PoissonStep, non-integer data type is forbidden.")
        elif np.any(self.data < 0):
            raise ValueError("Input data contains negative values.")

    def _init_moment(self):
        return BayesianPoissonMoment().init(self.data)

    def _continue(self, logbf):
        return np.log(self.skept) < logbf



import numpy as np
import matplotlib.pyplot as plt
from hmmlearn.utils import normalize
from scipy import special
from sfHMM._sfhmm_ext import forward, backward, viterbi, compute_log_xi_sum
from sfHMM.base import sfHMMBase


class sfHMMmotorBase(sfHMMBase):
    """Base class for sfHMM for motor stepping.

    This base class enables sparse transition probability matrix aiming at analyzing
    motor stepping trajectories. The attribute `transmat_` is generated from
    `transmat_kernel` every time it is called. Also, during M-step transmat_kernel is
    updated.
    """

    transmat_kernel: "np.ndarray"

    def __init__(
        self,
        sg0: float = -1,
        psf: float = -1,
        krange: "int | tuple[int, int] | None" = None,
        model: str = "g",
        name: str = "",
        max_stride: int = 2,
    ):
        super().__init__(
            sg0=sg0,
            psf=psf,
            krange=krange,
            model=model,
            name=name,
            covariance_type="tied",
        )
        self.max_stride = max_stride

    @property
    def transmat_(self):
        transmat = np.zeros((self.n_components, self.n_components))
        for i, p in enumerate(self.transmat_kernel):
            transmat += np.eye(self.n_components, k=i - self.max_stride) * p

        normalize(transmat, axis=1)
        return transmat

    def gmmfit(
        self,
        method: str = "Dirichlet",
        n_init: int = 1,
        random_state: int = 0,
        estimation: str = "fast",
    ):
        """
        estimation : str, optional
            How to estimate krange from step finding result.
            - "fast" ... narrower range but faster.
            - "safe" ... wider range but safer.
            or estimated krange can be directly given.
        """
        if self.krange is None:
            self._estimate_krange(estimation)
        return super().gmmfit(method, n_init, random_state)

    def tdp(self, **kwargs):
        dy_step = self._accumulate_step_sizes()
        dy_vit = np.array(
            [
                self.means_[sy] - self.means_[sx]
                for sx, sy in self.accumulate_transitions()
            ]
        )

        with plt.style.context(self.__class__.styles):
            if dy_step.size == 0:
                print("no step found")
                return None
            elif dy_vit.size > 0:
                xmin = min(dy_step.min(), dy_vit.min())
                xmax = max(dy_step.max(), dy_vit.max())
            else:
                xmin = dy_step.min()
                xmax = dy_step.max()
            plt.figure(figsize=(6, 4.4))
            # plot step sizes using step finding result
            plt.subplot(2, 1, 1)
            kw = dict(
                bins=int((self.max_stride * 2 + 1) * 5),
                color=self.__class__.colors["step finding"],
            )
            kw.update(kwargs)
            plt.hist(dy_step, **kw)
            plt.xlim(xmin, xmax)
            # plot step sizes using Viterbi path
            plt.subplot(2, 1, 2)
            kw = dict(
                bins=int((self.max_stride * 2 + 1) * 5),
                color=self.__class__.colors["Viterbi path"],
            )
            kw.update(kwargs)
            plt.hist(dy_vit, **kw)
            plt.xlim(xmin, xmax)

            plt.xlabel("step size")
            plt.show()
        return None

    def _init_sg0(self, p: float = 50):
        """Initialize 'sg0' if sg0 is negative."""
        return super()._init_sg0(p=p)

    def _check(self):
        self.startprob_ = np.asarray(self.startprob_)
        if len(self.startprob_) != self.n_components:
            raise ValueError("startprob_ must have length n_components")
        if not np.allclose(self.startprob_.sum(), 1.0):
            raise ValueError(
                f"startprob_ must sum to 1.0 (got {self.startprob_.sum():.4f})"
            )

        if not np.allclose(self.transmat_kernel.sum(), 1.0):
            raise ValueError(
                "transmat_kernel must sum to 1.0" f" (got {self.transmat_kernel.sum()})"
            )

    def _do_mstep(self, stats):
        if "s" in self.params:
            startprob_ = np.maximum(self.startprob_prior - 1 + stats["start"], 0)
            self.startprob_ = np.where(self.startprob_ == 0, 0, startprob_)
            normalize(self.startprob_)
        if "t" in self.params:
            transmat_ = np.maximum(self.transmat_prior - 1 + stats["trans"], 0)
            transmat_ = np.where(self.transmat_ == 0, 0, transmat_)
            for i in range(len(self.transmat_kernel)):
                self.transmat_kernel[i] = np.sum(
                    np.diag(transmat_, k=i - self.max_stride)
                )
            normalize(self.transmat_kernel)

        # GaussianHMM._do_mstep must be copied here.
        means_prior = self.means_prior
        means_weight = self.means_weight

        denom = stats["post"][:, None]
        if "m" in self.params:
            self.means_ = (means_weight * means_prior + stats["obs"]) / (
                means_weight + denom
            )

        if "c" in self.params:
            covars_prior = self.covars_prior
            covars_weight = self.covars_weight
            meandiff = self.means_ - means_prior

        if self.covariance_type in ("spherical", "diag"):
            c_n = (
                means_weight * meandiff**2
                + stats["obs**2"]
                - 2 * self.means_ * stats["obs"]
                + self.means_**2 * denom
            )
            c_d = max(covars_weight - 1, 0) + denom
            self._covars_ = (covars_prior + c_n) / np.maximum(c_d, 1e-5)
            if self.covariance_type == "spherical":
                self._covars_ = np.tile(
                    self._covars_.mean(1)[:, None], (1, self._covars_.shape[1])
                )
        elif self.covariance_type in ("tied", "full"):
            c_n = np.empty((self.n_components, self.n_features, self.n_features))
            for c in range(self.n_components):
                obsmean = np.outer(stats["obs"][c], self.means_[c])
                c_n[c] = (
                    means_weight * np.outer(meandiff[c], meandiff[c])
                    + stats["obs*obs.T"][c]
                    - obsmean
                    - obsmean.T
                    + np.outer(self.means_[c], self.means_[c]) * stats["post"][c]
                )
            cvweight = max(covars_weight - self.n_features, 0)
            if self.covariance_type == "tied":
                self._covars_ = (covars_prior + c_n.sum(axis=0)) / (
                    cvweight + stats["post"].sum()
                )
            elif self.covariance_type == "full":
                self._covars_ = (covars_prior + c_n) / (
                    cvweight + stats["post"][:, None, None]
                )

    def _do_viterbi_pass(self, framelogprob: np.ndarray):
        n_samples, n_components = framelogprob.shape
        state_sequence, logprob = viterbi(
            n_samples,
            n_components,
            log_mask_zero(self.startprob_),
            log_mask_zero(self.transmat_kernel),
            framelogprob,
            self.max_stride,
        )
        return logprob, state_sequence

    def _do_forward_pass(self, framelogprob: np.ndarray):
        n_samples, n_components = framelogprob.shape
        fwdlattice = forward(
            n_samples,
            n_components,
            log_mask_zero(self.startprob_),
            log_mask_zero(self.transmat_kernel),
            framelogprob,
            self.max_stride,
        )
        with np.errstate(under="ignore"):
            return special.logsumexp(fwdlattice[-1]), fwdlattice

    def _do_backward_pass(self, framelogprob: np.ndarray):
        n_samples, n_components = framelogprob.shape
        bwdlattice = backward(
            n_samples,
            n_components,
            log_mask_zero(self.transmat_kernel),
            framelogprob,
            self.max_stride,
        )
        return bwdlattice

    def _accumulate_sufficient_statistics(
        self, stats, obs, framelogprob, posteriors, fwdlattice, bwdlattice
    ):
        stats["nobs"] += 1
        if "s" in self.params:
            stats["start"] += posteriors[0]
        if "t" in self.params:
            n_samples, n_components = framelogprob.shape
            # when the sample is of length 1, it contains no transitions
            # so there is no reason to update our trans. matrix estimate
            if n_samples <= 1:
                return

            log_xi_sum = compute_log_xi_sum(
                n_samples,
                n_components,
                fwdlattice,
                log_mask_zero(self.transmat_kernel),
                bwdlattice,
                framelogprob,
                self.max_stride,
            )
            with np.errstate(under="ignore"):
                stats["trans"] += np.exp(log_xi_sum)

        if "m" in self.params or "c" in self.params:
            stats["post"] += posteriors.sum(axis=0)
            stats["obs"] += np.dot(posteriors.T, obs)

        if "c" in self.params:
            if self.covariance_type in ("spherical", "diag"):
                stats["obs**2"] += np.dot(posteriors.T, obs**2)
            elif self.covariance_type in ("tied", "full"):
                # posteriors: (nt, nc); obs: (nt, nf); obs: (nt, nf)
                # -> (nc, nf, nf)
                stats["obs*obs.T"] += np.einsum("ij,ik,il->jkl", posteriors, obs, obs)

    def _get_n_fit_scalars_per_param(self):
        nc = self.n_components
        nf = self.n_features
        return {
            "s": nc - 1,
            "t": self.max_stride * 2 + 1,
            "m": nc * nf,
            "c": {
                "spherical": nc,
                "diag": nc * nf,
                "full": nc * nf * (nf + 1) // 2,
                "tied": nf * (nf + 1) // 2,
            }[self.covariance_type],
        }


def log_mask_zero(a):
    with np.errstate(divide="ignore"):
        return np.log(a)



from numpy import ndarray

def forward(
    n_samples: int,
    n_components: int,
    log_startprob: ndarray,
    log_transmat_kernel: ndarray,
    framelogprob: ndarray,
    max_stride: int,
) -> ndarray: ...
def backward(
    n_samples: int,
    n_components: int,
    log_transmat_kernel: ndarray,
    framelogprob: ndarray,
    max_stride: int,
) -> ndarray: ...
def compute_log_xi_sum(
    n_samples: int,
    n_components: int,
    fwdlattice: ndarray,
    log_transmat_kernel: ndarray,
    bwdlattice: ndarray,
    framelogprob: ndarray,
    max_stride: int,
) -> ndarray: ...
def viterbi(
    n_samples: int,
    n_components: int,
    log_startprob: ndarray,
    log_transmat_kernel: ndarray,
    framelogprob: ndarray,
    max_stride: int,
) -> tuple[ndarray, float]: ...



def check_ref(ref, class_obj):
    if ref is None:
        ref = class_obj()
    elif not isinstance(ref, class_obj):
        raise TypeError(
            f"`out` must be an object of {class_obj.__name__} or its subclass."
        )
    return ref


def infer_sep(path: str, encoding: str = None):
    if path.endswith(".csv"):
        sep = ","
    elif path.endswith(".tsv"):
        sep = "\t"
    elif path.endswith(".dat"):
        sep = r"\s+"
    else:
        with open(path, encoding=encoding) as f:
            line0 = f.readline()
            line1 = f.readline()
        if line0.count(",") == line1.count(",") > 0:
            sep = ","
        elif line0.count("\t") == line1.count("\t") > 0:
            sep = "\t"
        else:
            sep = r"\s+"

    return sep



import numpy as np
from scipy.special import softmax
from scipy.stats import entropy
from scipy.optimize import minimize
import matplotlib.pyplot as plt
from hmmlearn.utils import normalize
from functools import wraps
from warnings import warn


class sfHMMAnalysisError(Exception):
    pass


def append_log(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if len(self._log) > 0 and self._log[-1][0] == func.__name__:
            self._log[-1][1] += 1
            self._log[-1][2] = "Failed"
        else:
            self._log.append([func.__name__, 1, "Failed"])
        out = func(self, *args, **kwargs)
        self._log[-1][2] = "Passed"
        return out

    return wrapper


def under_development(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        msg = f"Method `{func.__name__}` is under development and its behavior may change in the future."
        warn(msg, FutureWarning, stacklevel=2)
        return func(self, *args, **kwargs)

    return wrapper


def gauss_mix(x, gmm):
    return np.exp(gmm.score_samples(x.reshape(-1, 1)))


def concat(list_of_list):
    out = []
    for list_ in list_of_list:
        out += list(list_)
    return out


def plot2(data1, data2=None, ylim=None, legend=True, color1=None, **kwargs):
    """
    dict_ = {"color": ___, "label": ___}
    """
    if ylim is None:
        ylim = [np.min(data1), np.max(data1)]

    plt.xlim(0, len(data1))
    plt.ylim(ylim)

    plt.plot(data1, color=color1, label="raw data")

    if data2 is not None:
        plt.plot(data2, **kwargs)

    if legend:
        plt.legend(bbox_to_anchor=(1.05, 1), loc="upper left", borderaxespad=0)

    return None


def calc_covars(data_raw, states, n_components):
    if n_components <= 1:
        return np.array([np.var(data_raw)])

    covars = np.empty(n_components)

    for s in range(n_components):
        index = states == s
        if np.sum(index) > 0:
            covars[s] = np.var(data_raw[index])
            if covars[s] == 0:
                covars[s] = np.var(data_raw)
        else:
            covars[s] = np.var(data_raw)

    return covars


def calc_startprob(d0_list, gmm):
    # Here, using `scipy.special.softmax` is very important because logprob always
    # composed of very large and very small values. This situation causes underflow
    # or overflow.
    logprob = gmm._estimate_weighted_log_prob(np.asarray(d0_list).reshape(-1, 1))
    sumlogprob = np.sum(logprob, axis=0)
    return softmax(sumlogprob)


def calc_transmat(states_list, n_components):
    transmat = np.zeros((n_components, n_components), dtype=np.float64)

    for states in states_list:
        for i in range(len(states) - 1):
            transmat[states[i], states[i + 1]] += 1.0
    transmat += 1e-12
    normalize(transmat, axis=1)
    return transmat


def normalized_mutual_information(a1, a2, bins, range):
    # (H(X) + H(Y)) / H(X,Y)
    a1, _ = np.histogram(a1, bins=bins, range=range, density=True)
    a2, _ = np.histogram(a2, bins=bins, range=range, density=True)
    if np.isnan(a2).any():
        return 0
    hist, _ = np.histogramdd([a1, a2], bins=bins)
    hist /= np.sum(hist)
    h1 = entropy(np.sum(hist, axis=0))
    h2 = entropy(np.sum(hist, axis=1))
    h12 = entropy(np.ravel(hist))

    return (h1 + h2) / h12


def optimize_ax(d1, d2, bins=None, range=None, bounds=None):
    def calc_nmi(a, d1, d2):
        return -normalized_mutual_information(d1, a * d2, bins, range)

    result = minimize(calc_nmi, 1, args=(d1, d2), method="Powell", bounds=bounds)
    return result.x, 0


def optimize_b(d1, d2, bins=None, range=None, bounds=None):
    def calc_nmi(b, d1, d2):
        return -normalized_mutual_information(d1, d2 + b, bins, range)

    result = minimize(calc_nmi, 0, args=(d1, d2), method="Powell", bounds=bounds)
    return 1, result.x



from __future__ import annotations

import numpy as np
from sklearn import mixture
import matplotlib.pyplot as plt
from .utils import gauss_mix


class GMM1(mixture.GaussianMixture):
    """
    GaussianMixture with sorted parameters.
    """

    def __init__(self, n_components, **kwargs):
        super().__init__(n_components=n_components, **kwargs)
        self.valid = False

    def fit(self, data, scale=1.0):
        super().fit(data / scale)

        # sort all
        order = np.argsort(self.means_.flat)
        self.weights_ = self.weights_[order]
        self.means_ = self.means_[order]
        if self.covariance_type == "spherical":
            self.covariances_ = self.covariances_[order]
            self.precisions_cholesky_ = self.precisions_cholesky_[order]
            self.precisions_ = self.precisions_[order]

        self.sigma_ = np.sqrt(self.covariances_.flat)
        self.valid = True
        return self


class GMMs:
    """
    Gaussian mixture models with different states.
    The best model can be chosen by comparing AIC or BIC.
    """

    def __init__(
        self, data, krange, covariance_type="spherical", min_interval=None, min_sg=None
    ):
        """
        Parameters
        ----------
        data : 1D array
            Raw data for clustering.
        krange : (int, int)
            Range of number of states. (2, 6) means states 2 <= k <= 6 will be tested
        covariance_type : str, by default "spherical"
            See GaussianMixture.
        min_interval : float, optional
            Fitting results with interval(s) of peaks smaller than this will be
            considered as invalid fitting.
        min_sg : float, optional
            Fitting results with standard deviation(s) smaller than this will be
            considered as invalid fitting.
        """
        self.data = data
        self.klist = list(range(krange[0], krange[1] + 1))
        self.results = None
        self.covariance_type = covariance_type
        self.min_interval = min_interval if min_interval else -1
        self.min_sg = min_sg if min_sg else -1

    def __getitem__(self, k):
        """
        self[k] returns k-state model.
        """
        return self.results[k]

    def __repr__(self):
        out = "Gaussian Mixture Models"
        line1 = "  n  ||"
        line2 = " AIC ||"
        line3 = " BIC ||"
        for n, aic, bic in zip(self.klist, self.get_aic(), self.get_bic()):
            line1 += f" {n:>7} |"
            line2 += f" {int(aic+0.5):>7} |"
            line3 += f" {int(bic+0.5):>7} |"
        return "\n".join([out, line1, line2, line3])

    def show_aic_bic(self):
        beautiful_red = "#EE08A1"
        beautiful_blue = "#0818E6"
        y_aic = self.get_aic()
        y_bic = self.get_bic()
        x = self.klist
        plt.figure(figsize=(4.5, 3.2))
        plt.plot(x, y_aic, color=beautiful_red, marker="D", label="AIC")
        plt.plot(x, y_bic, color=beautiful_blue, marker="D", label="BIC")
        plt.legend()
        plt.xlabel("number of states")
        return None

    def plot_all(self):
        x = np.linspace(self.data.min(), self.data.max(), 256)
        n_bin = min(int(np.sqrt(self.data.size * 2)), 256)
        for n, gmm in self.results.items():
            plt.figure(figsize=(4.0, 1.6))
            y = gauss_mix(x, gmm)
            plt.hist(
                self.data, bins=n_bin, color="lightgray", density=True, label="raw data"
            )
            plt.plot(x, y, color="red", label="fit")
            plt.title(f"number of states = {n}")
            plt.legend(bbox_to_anchor=(1.05, 1), loc="upper left", borderaxespad=0)
            plt.show()
        return None

    def fit(self, n_init: int = 1, random_state: int = 0, scale: float = 1.0, **kwargs):
        """
        Fit data to k-state model, for every k in self.klist. Peak intervals and
        standard deviations are checked for every fitting.

        Parameters
        ----------
        n_init : int, default is 1
            Number of initialization (Kmeans method depends on random initial values).
        random_state : int, default is 0
            Random state used in Kmeans initialization.
        scale : float, default is 1.0
            Normalization factor to make fitting scalable.
        """
        scale = 1.0 if scale <= 0 else scale
        d = np.asarray(self.data).reshape(-1, 1) / scale

        self.results = {
            k: GMM1(
                k,
                n_init=n_init,
                random_state=random_state,
                covariance_type=self.covariance_type,
                **kwargs,
            )
            for k in self.klist
        }

        for gmm1 in self.results.values():
            gmm1.fit(d)
            # rescale
            gmm1.means_ *= scale
            gmm1.covariances_ *= scale**2
            gmm1.precisions_cholesky_ /= scale
            gmm1.precisions_ /= scale**2
            gmm1.sigma_ *= scale

            if self._interval_check(gmm1.means_) or self._sg_check(gmm1.sigma_):
                gmm1.valid = False

        return self

    def get_optimal(self, criterion: str = "bic", only_valid: bool = True):
        """
        Determine the best model that minimize AIC or BIC.

        Parameters
        ----------
        criterion : str, by default "bic"
            Which criterion is used.
        only_valid : bool, by default True
            If True, only valid results are considered.

        Returns
        -------
        GMM1 object
            The best model
        """
        if criterion == "bic":
            cri_list = self.get_bic()
        elif criterion == "aic":
            cri_list = self.get_aic()
        else:
            raise ValueError("'criterion' must be either 'aic' or 'bic'.")

        if only_valid:
            ng = self.isvalid()
            if ng.all():
                pass
            else:
                cri_list[~self.isvalid()] = np.inf

        k_best = self.klist[np.argmin(cri_list)]
        return self[k_best]

    def get_aic(self):
        """
        Calculate Akaike's Information Criterion (AIC) for every state.

        Returns
        -------
        list of float
        """
        d = np.asarray(self.data).reshape(-1, 1)
        return np.array([self[k].aic(d) for k in self.klist])

    def get_bic(self):
        """
        Calculate Bayesian Information Criterion (BIC) for every state.

        Returns
        -------
        list of float
        """
        d = np.asarray(self.data).reshape(-1, 1)
        return np.array([self[k].bic(d) for k in self.klist])

    def isvalid(self):
        """
        Returns if fittings are valid.

        Returns
        -------
        list of bool
        """
        return np.array([gmm1.valid for gmm1 in self.results.values()])

    def _interval_check(self, mu):
        """
        check if peaks are too near to each other.
        """
        return (np.diff(mu.flat) < self.min_interval).any()

    def _sg_check(self, sg):
        """
        check if any standard deviations are too small.
        """
        return (sg < self.min_sg).any()


class DPGMM(mixture.BayesianGaussianMixture):
    def __init__(self, n_components, n_init, random_state, **kwargs):
        # change default
        kw = dict(
            weight_concentration_prior=1,
            weight_concentration_prior_type="dirichlet_distribution",
            covariance_type="spherical",
            max_iter=1000,
        )
        kw.update(kwargs)

        # different covariance types require different input
        if "covariance_prior" in kw:
            c = kw["covariance_prior"]
            if kw["covariance_type"] == "tied" and isinstance(c, (int, float)):
                kw["covariance_prior"] = [[c]]
            elif kw["covariance_type"] == "spherical" and isinstance(c, np.ndarray):
                kw["covariance_prior"] = float(c)

        super().__init__(
            n_components=n_components, n_init=n_init, random_state=random_state, **kw
        )

    def fit(self, data, scale: float = 1.0):
        data = data / scale
        if self.mean_precision_prior is None:
            self.mean_precision_prior = 1 / np.var(data)
        super().fit(data)

        labels = self.predict(data)
        unique_labels = np.unique(labels)
        self.n_components = len(unique_labels)

        # sort all
        self.means_ = self.means_[unique_labels] * scale
        order = np.argsort(self.means_.flat)
        self.weights_ = self.weights_[unique_labels][order]
        self.means_ = self.means_[order]
        if self.covariance_type == "spherical":
            self.covariances_ = self.covariances_[unique_labels][order]
            self.precisions_cholesky_ = self.precisions_cholesky_[unique_labels][order]
            self.precisions_ = self.precisions_[unique_labels][order]
            self.degrees_of_freedom_ = self.degrees_of_freedom_[unique_labels][order]

        if self.weight_concentration_prior_type == "dirichlet_process":
            self.weight_concentration_ = (
                self.weight_concentration_[0][unique_labels][order],
                self.weight_concentration_[1][unique_labels][order],
            )
        else:
            self.weight_concentration_ = self.weight_concentration_[unique_labels][
                order
            ]

        # rescale
        self.covariances_ *= scale**2
        self.precisions_cholesky_ /= scale
        self.precisions_ /= scale

        self.mean_precision_ = self.mean_precision_[unique_labels][order]
        self.sigma_ = np.sqrt(self.covariances_.flat)

        return self



from __future__ import annotations

import re
import copy
from typing import Callable, Iterable, Iterator, Literal, overload, TYPE_CHECKING
import numpy as np
import matplotlib.pyplot as plt
from .utils import (
    append_log,
    plot2,
    concat,
    calc_covars,
    calc_startprob,
    calc_transmat,
    sfHMMAnalysisError,
    optimize_ax,
    optimize_b,
    under_development,
)
from .single_sfhmm import sfHMM1, _S
from .base import sfHMMBase

if TYPE_CHECKING:
    import pandas as pd

__all__ = ["sfHMMn"]


class sfHMMn(sfHMMBase):
    r"""
    Multi-trajectory sfHMM.
    This class shares many attributes in sfHMM1.

    Analysis Results
    ----------------
    gmm_opt : `GMMs` or `DPGMM` object. For more detail, see `sfHMM.gmm`.
        The result of GMM clustering, which has following attributes:
        - wt ... Weights.
        - mu ... Means.
        - sg ... Standard deviations.
        - n_components ... The number of states.
        If AIC/BIC minimization of standard GMM clustering was conducted, the clustering
        results will be stored in 'gmm'.
    n_components : int
        The optimal number of states. Same as 'gmm_opt.n_components'.
    self[i] (accessed by indexing) : sfHMM object
        The i-th sfHMM object. See .\single_sfhmm.py.
    """

    def __init__(
        self,
        data_raw: Iterable[_S] | None = None,
        *,
        sg0: float = -1,
        psf: float = -1,
        krange: int | tuple[int, int] = None,
        model: str = "g",
        name: str = "",
        **kwargs,
    ):
        """
        Parameters
        ----------
        data_raw : iterable, optional
            Datasets to be added.
        sg0 : float, optional
            Parameter used in filtering method. Expected to be 20% of signal change.
            If <= 0, sg0 will be determined automatically.
        psf : float, optional
            Transition probability used in step finding algorithm.
            if 0 < p < 0.5 is not satisfied, the original Kalafut-Visscher's algorithm
            is executed.
        krange : int or (int, int)
            Minimum and maximum number of states to search in GMM clustering. If it is
            integer, then
            it will be interpretted as (krange, krange).
        model: str, by default "g" (= Gaussian)
            Distribution of noise. Gauss and Poisson distribution are available for now.
        name : str, optional
            Name of the object.
        """
        self.n_data = 0
        super().__init__(sg0, psf, krange, model, name, **kwargs)
        self.ylim = [np.inf, -np.inf]
        self._sf_list = []
        self.gmm_opt = None
        data_raw is None or self.appendn(data_raw)

    @overload
    def __getitem__(self, key: int) -> sfHMM1: ...
    @overload
    def __getitem__(self, key: slice) -> sfHMMn: ...

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._sf_list[key]
        elif isinstance(key, slice):
            _k = ":".join([str(key.start), str(key.stop), str(key.step or "")])
            if _k.endswith(":"):
                _k = _k[:-1]
            new = self.__class__(
                sg0=self.sg0,
                psf=self.psf,
                krange=self.krange,
                model=self.model,
                name=f"{self.name}[{_k}]",
            )
            new._sf_list = self._sf_list[key]
            new.ylim = self.ylim
            new.n_data = len(new._sf_list)
            return new
        else:
            raise TypeError(f"Index key must be int or slice, got {type(key)}")

    def __iter__(self) -> Iterator[sfHMM1]:
        return iter(self._sf_list)

    def __add__(self, other: sfHMMn) -> sfHMMn:
        """
        `+` is supported for two sfHMMn objects. a+b makes a new sfHMMn object with
        concatenated list of sfHMM1 objects.
        """
        if self is other:
            new = self.__class__(
                sg0=self.sg0,
                psf=self.psf,
                krage=self.krange,
                model=self.model,
                name=self.name + "+" + other.name,
            )
            new.n_data = self.n_data + other.n_data
            new.ylim = [
                min(self.ylim[0], other.ylim[0]),
                max(self.ylim[1], other.ylim[1]),
            ]
            new._sf_list = copy.deepcopy(self._sf_list) + copy.deepcopy(other._sf_list)
            return new
        else:
            raise TypeError(
                "Unsupported operand type(s) for +: "
                f"{self.__class__.__name__} and {type(other)}."
            )

    @property
    def names(self) -> list[str]:
        return [sf.name for sf in self]

    @append_log
    def append(self, data: _S, name: str = None) -> sfHMMn:
        """
        Append a trajectory as sfHMM object.

        Parameters
        ----------
        data : array
            Data to analyze.
        name : str, optional
            Name of the data.
        """
        if name is None:
            name = self.name + f"[{self.n_data}]"

        sf = sfHMM1(
            data,
            sg0=self.sg0,
            psf=self.psf,
            krange=self.krange,
            model=self.StepClass,
            name=name,
        )
        self.n_data += 1
        self._sf_list.append(sf)
        self.ylim[0] = min(sf.ylim[0], self.ylim[0])
        self.ylim[1] = max(sf.ylim[1], self.ylim[1])
        return self

    def appendn(self, datasets: Iterable[_S]) -> sfHMMn:
        """Append all the data in `datasets`.

        Parameters
        ----------
        datasets : dict, list or iterable objects except for np.ndarray or pd.DataFrame.
            Datasets to be appended. If it is dict, then keys are interpreted as names.
        """
        import pandas as pd

        if isinstance(datasets, dict):
            self.from_dict(datasets)
        elif isinstance(datasets, np.ndarray):
            raise TypeError(
                "Datasets of ndarray is ambiguious. Please use self.append(a) for 1-D "
                "ndarray, or explicitly specify along which axis to iterate by such as "
                "list(a) or np.split(a, axis=1)."
            )
        elif isinstance(datasets, pd.DataFrame):
            self.from_pandas(datasets)
        else:
            for data in datasets:
                self.append(data)

        return self

    @append_log
    def delete(self, indices: int | Iterable[int]) -> None:
        """
        Delete sfHMM1 object(s) from the list.

        Parameters
        ----------
        indices : int or iterable of int
            Indices to delete.
        """
        if isinstance(indices, int):
            indices = [indices]
        indices = sorted(indices)

        for i in reversed(indices):
            self._sf_list.pop(i)

        data_raw_all = self.data_raw
        self.ylim = [data_raw_all.min(), data_raw_all.max()]
        self.n_data -= len(indices)

        return None

    def deleteif(self, filter_func: Callable[[sfHMM1], bool]) -> list[int]:
        """
        Delete sfHMM1 object(s) from the list if certain conditions are
        satisfied.

        Parameters
        ----------
        filter_func : callable
            `filter_func(sf)==True` then delete `sf`.

        Returns
        -------
        list of int
            Deleted indices.
        """
        if not callable(filter_func):
            raise TypeError("`filter_func` must be callable")

        indices = [i for i, sf in enumerate(self) if filter_func(sf)]

        len(indices) > 0 and self.delete(indices)

        return indices

    @append_log
    def pop(self, ind: int) -> sfHMM1:
        """
        Delete one sfHMM1 object and return it.

        Parameters
        ----------
        ind : int
            Indice to pop.
        """
        out = self._sf_list.pop(ind)
        data_raw_all = self.data_raw
        self.ylim = [data_raw_all.min(), data_raw_all.max()]
        self.n_data -= 1
        return out

    def from_dict(self, d: dict[str, _S], like: str = None, regex: str = None):
        """
        Load datasets from dict.

        Parameters
        ----------
        d : dict
            Input data.
        like : str, optional
            If given, key that contains this string is appended.
        regex : regular expression, optional
            If given, key that matches this regular expression is appended.
        """
        for name, data in d.items():
            # only keys matched like or regex requirement are appended.
            if like and like not in name:
                continue
            if regex and not re.match(regex, name):
                continue

            self.append(data, name=str(name))

        if self.n_data == 0:
            raise ValueError(
                "No data appended. Confirm that input dict is in a correct format."
            )

        return self

    def from_pandas(
        self,
        df: pd.DataFrame,
        like: str | None = None,
        regex: str | None = None,
        melt: bool | Literal["infer"] = "infer",
    ) -> sfHMMn:
        """
        Load datasets from pandas.DataFrame.

        Parameters
        ----------
        df : DataFrame
            Input data.
        like : str, optional
            If given, dataset that contains this string is appended.
        regex : regular expression, optional
            If given, dataset that matches this regular expression is appended.
        melt : bool or "infer", optional
            If input DataFrame is melted, which is automatically determined when melt is
            "infer".
        """
        if melt == "infer":
            # Determine if df is melted or not
            col0 = df.iloc[:, 0]
            if df.shape[1] == 2 and len(col0.unique()) < len(col0) // 20:
                return self.from_pandas(df, like, regex, melt=True)
            else:
                return self.from_pandas(df, like, regex, melt=False)

        if melt:
            if df.shape[1] != 2:
                raise ValueError(
                    "For melted DataFrame, it must composed of two columns, with names "
                    "in the first and values in the second."
                )
            name_col, value_col = df.columns
            for name in df[name_col].unique():
                if like and like not in name:
                    continue
                elif regex and not re.match(regex, name):
                    continue
                self.append(df[df[name_col] == name][value_col], name)

        else:
            if like is not None or regex is not None:
                df = df.filter(like=like, regex=regex)
            for name in df:
                data = df[name].dropna()  # delete NaN
                self.append(data, name)  # append data

        if self.n_data == 0:
            raise ValueError(
                "No data appended. Confirm that input DataFrame is in a correct format."
            )

        return self

    def read(
        self,
        path: str,
        sep: str = None,
        encoding: str = None,
        header: int = 0,
        **kwargs,
    ):
        """
        Read a file using appropriate function, and import its data to sfHMMn object.

        Parameters
        ----------
        path : str
            Path to file.
        sep, encoding, header
            Important arguments in pd.read_csv. Default is header=0 rather than "infer"
            because header="infer" usually works in a wrong way.
        **kwargs
            Other keyword arguments that will passed to pd.read_csv or pd.read_excel.
        """
        from .io import read

        read(path, out=self, sep=sep, encoding=encoding, header=header, **kwargs)
        self.source = path
        return self

    @append_log
    def step_finding(self) -> sfHMMn:
        """
        Step finding by extended version of Kalafut-Visscher's algorithm.
        Run independently for each sfHMM object.
        """
        if self.n_data <= 0:
            raise sfHMMAnalysisError("Cannot start analysis before appending data.")

        for sf in self:
            sf.psf = self.psf
            sf.step_finding()
        return self

    @append_log
    def denoising(self) -> sfHMMn:
        """
        Denoising by cutting of the standard deviation of noise to sg0.
        Run independently for each sfHMM object.
        """
        if self.n_data <= 0:
            raise sfHMMAnalysisError("Cannot start analysis before appending data.")

        self._init_sg0()

        for sf in self:
            sf.sg0 = self.sg0
            sf.denoising()

        return self

    @append_log
    def gmmfit(
        self, method: str = "bic", n_init: int = 1, random_state: int = 0
    ) -> sfHMMn:
        """
        Fit the denoised data to Gaussian mixture model.

        Paramters
        ---------
        method : str, optional
            How to determine the optimal number of states. This parameter must be
            'aic', 'bic' or 'Dirichlet'. By default "bic".
        n_init : int, optional
            How many times initialization will be performed in K-means, by default 1.
        random_state : int , optional
            Random seed for K-means initialization., by default 0.
        """
        if self.n_data <= 0:
            raise sfHMMAnalysisError("Cannot start analysis before appending data.")

        self._gmmfit(method, n_init, random_state)

        for sf in self:
            sf.states = self.gmm_opt.predict(np.asarray(sf.step.fit).reshape(-1, 1))
            sf.n_components = self.n_components
        return self

    @append_log
    def hmmfit(self) -> sfHMMn:
        """
        HMM paramter optimization by Forward-Backward algorithm, and state inference by
        Viterbi algorithm.
        """
        if self.n_data <= 0:
            raise sfHMMAnalysisError("Cannot start analysis before appending data.")

        self.data_raw_all = self.data_raw
        self.states_list = [sf.states for sf in self]

        self._set_hmm_params()

        self.fit(self.data_raw_all.reshape(-1, 1), lengths=self.size)

        for sf in self:
            self._copy_params(sf)
            sf.states = sf.predict(np.asarray(sf.data_raw).reshape(-1, 1))
            sf.viterbi = sf.means_[sf.states, 0]
        del self.data_raw_all, self.states_list
        return self

    def run_all_separately(self, plot_every: int = 0) -> sfHMMn:
        """
        Run the function `run_all` for every sfHMM1 object.
        Paramters
        ---------
        plot_every : int
            Every `plot_every` result will be plotted during iteration.
        """
        for i, sf in enumerate(self):
            pl = plot_every > 0 and i % plot_every == 0
            try:
                sf.run_all(plot=pl)
            except Exception as e:
                print(f"{e.__class__.name} during {i}-th trace: {e}")

        return self

    def _set_covars(self):
        self.covars_ = calc_covars(
            self.data_raw_all,
            concat(self.states_list),
            self.n_components,
        )
        return None

    def _set_means(self):
        self.means_ = self.gmm_opt.means_.copy()
        return None

    def _set_startprob(self):
        d0_list = [sf.data_raw[0] for sf in self]
        self.startprob_ = calc_startprob(d0_list, self.gmm_opt)
        return None

    def _set_transmat(self):
        self.transmat_ = calc_transmat(self.states_list, self.n_components)
        return None

    def plot(self):
        if self.gmm_opt is not None:
            self.plot_hist()
        self.plot_traces()
        return None

    def plot_hist(self):
        """
        Plot histogram of data_raw, data_fil and GMM fitting result.
        """
        with plt.style.context(self.__class__.styles):
            plt.figure(figsize=(3, 4))
            plt.suptitle(self.name, fontweight="bold")
            self._hist()
            plt.show()
        return None

    def plot_traces(
        self,
        data: str = "Viterbi path",
        filter_func: Callable | None = None,
        sharex: bool = False,
    ):
        """
        Plot all the trajectories. The figure will look like:
         __ __ __ __
        | 0| 1| 2| 3|
        |__|__|__|__|
        | 4| 5| 6| 7|
        |__|__|__|__|
        | 8| 9|10|11|
        :     :     :

        Parameters
        ----------
        data : str, default is "Viterbi path"
            Which data to plot over the raw data trajectories.
        filter_func : callable or None, optional
            If not None, only sfHMM objects that satisfy filter_func(sf)==True are
            plotted.
        sharex : bool, default is False
            If True, all the subplots will share x-limits.
        """
        c_other = self.colors.get(data, None)

        # index list that satisfies filter_func
        if filter_func is None:
            indices = range(self.n_data)
        else:
            indices = [i for (i, sf) in enumerate(self) if filter_func(sf)]

        n_col = 4
        n_row = (len(indices) - 1) // n_col + 1

        if sharex:
            xlim = [0, max(sf.size for sf in self)]

        with plt.style.context(self.__class__.styles):
            plt.figure(figsize=(n_col * 2.9, n_row * 2.4))

            for i, ind in enumerate(indices):
                sf = self[ind]
                ax = plt.subplot(n_row, n_col, i + 1)

                if data == "Viterbi path":
                    d = sf.viterbi
                elif data == "denoised":
                    d = sf.data_fil
                elif data == "step finding":
                    d = sf.step.fit
                elif data == "none":
                    d = None
                else:
                    raise ValueError(
                        "'data' must be 'step finding', 'denoised', "
                        "'Viterbi path' or 'none'"
                    )

                plot2(
                    sf.data_raw,
                    d,
                    ylim=self.ylim,
                    legend=False,
                    color1=self.colors["raw data"],
                    color=c_other,
                )

                sharex and plt.xlim(xlim)

                plt.text(
                    ax.get_xlim()[1] * 0.98,
                    ax.get_ylim()[1] * 0.98,
                    str(ind),
                    ha="right",
                    va="top",
                    color="gray",
                )

            plt.tight_layout()
            plt.show()
        return None

    def view_in_qt(self, title: str | None = None):
        """
        Open a Qt viewer and plot the results.
        """
        from .viewer import TrajectoryViewer

        if title is None:
            title = self.name

        datalist = []
        for sf in self:
            data = dict()
            sf.data_raw is None or data.update({"raw data": sf.data_raw})
            sf.step is None or data.update({"step finding": sf.step.fit})
            sf.data_fil is None or data.update({"denoised": sf.data_fil})
            sf.viterbi is None or data.update({"Viterbi path": sf.viterbi})
            datalist.append(data)

        viewer = TrajectoryViewer(
            datalist, self.__class__.styles, self.__class__.colors
        )
        viewer.setWindowTitle(title)
        viewer.show()
        return viewer

    def accumulate_transitions(self) -> list[tuple[int, int]]:
        return concat([sf.accumulate_transitions() for sf in self])

    def _accumulate_step_sizes(self) -> np.ndarray:
        if self[0].step is None:
            raise sfHMMAnalysisError("Steps are not detected yet.")
        return np.array(concat([sf.step.step_size_list for sf in self]))

    def _copy_params(self, sf: sfHMM1):
        if self.covariance_type == "spherical":
            sf.covars_ = self.covars_.ravel()
        else:
            sf.covars_ = [[self.covars_[0, 0, 0]]]
        sf.min_covar = self.min_covar
        sf.means_ = self.means_
        sf.startprob_ = self.startprob_
        sf.transmat_ = self.transmat_
        return None

    @under_development
    @append_log
    def align(
        self,
        bounds: tuple[float, float],
        bins: int = 32,
        formula: Literal["ax", "y=ax", "b", "y=x+b"] = "y=ax",
    ) -> list:
        """
        Align step finding results with `formula` transformation. The optimal parameter
        is determined by minimizing normalized mutual information of two step finding
        results: `self[0].step.fit` as the reference and `a * self[i].step.fit + b` as
        the variable.

        (1) y = ax
          __                  __    __
         |      __           |     |  |
        _|    _|  |__  -->  _|    _|  |__

        (2) y = x + b
        _                  _       __
         |__    __          |__  _|  |__
              _|  |__  -->


        Parameters
        ----------
        bounds : tuple of floats
            Bounds of parameter $a$ or $b$, i.e. optimal parameter is searched in the
            range from `bounds[0]` to `bounds[1]`.
        bins : int, default is 32
            Bin number for calculating shannon entropy and mutual information.
        formula: str, default is "y=ax"
            Formulation of transformation.

        """

        if self.n_data < 2:
            raise sfHMMAnalysisError("Cannot align datasets because n_data < 2.")
        if self[0].step is None:
            raise sfHMMAnalysisError("Cannot align datasets before step finding.")

        formula = re.sub(" ", "", formula)
        optimization_func = {
            "ax": optimize_ax,
            "y=ax": optimize_ax,
            "b": optimize_b,
            "y=x+b": optimize_b,
        }[formula]

        for sf in self[1:]:
            result = optimization_func(
                self[0].step.fit,
                sf.step.fit,
                bins=bins,
                range=self.ylim,
                bounds=[bounds],
            )

            a, b = result
            sf.data_raw[:] = a * sf.data_raw + b
            sf.ylim = [a * sf.ylim[0] + b, a * sf.ylim[1] + b]
            sf.step.fit[:] = a * sf.step.fit + b
            sf.step.mu_list[:] = a * sf.step.mu_list + b
            sf.step.step_size_list[:] = a * sf.step.step_size_list
            if sf.data_fil is not None:
                sf.data_fil[:] = a * sf.data_fil + b

        return None

    @property
    def data_raw(self) -> np.ndarray:
        return np.array(concat([sf.data_raw for sf in self]))

    @property
    def data_fil(self) -> np.ndarray:
        return np.array(concat([sf.data_fil for sf in self]))

    @property
    def _sg_list(self) -> np.ndarray:
        return np.array(concat([sf._sg_list for sf in self]))

    @property
    def size(self) -> list[int]:
        return [sf.size for sf in self]



from matplotlib.backends.backend_qt5agg import FigureCanvas
from matplotlib.backend_bases import MouseEvent, MouseButton

__all__ = ["EventedCanvas"]


class EventedCanvas(FigureCanvas):
    """
    A figure canvas implemented with mouse callbacks.
    """

    def __init__(self, fig):
        super().__init__(fig)
        self.pressed = None
        self.lastx = None
        self.lasty = None

    def wheelEvent(self, event):
        """
        Resize figure by changing axes xlim and ylim. If there are subplots, only the subplot
        in which cursor exists will be resized.
        """
        fig = self.figure

        delta = event.angleDelta().y() / 120
        event = self.get_mouse_event(event)

        if not event.inaxes:
            return None
        for ax in fig.axes:
            if event.inaxes != ax:
                continue
            x0, x1 = ax.get_xlim()
            y0, y1 = ax.get_ylim()

            if delta > 0:
                factor = 1 / 1.3
            else:
                factor = 1.3

            ax.set_xlim(
                [
                    (x1 + x0) / 2 - (x1 - x0) / 2 * factor,
                    (x1 + x0) / 2 + (x1 - x0) / 2 * factor,
                ]
            )
            ax.set_ylim(
                [
                    (y1 + y0) / 2 - (y1 - y0) / 2 * factor,
                    (y1 + y0) / 2 + (y1 - y0) / 2 * factor,
                ]
            )
            break

        fig.canvas.draw()
        return None

    def mousePressEvent(self, event):
        """
        Record the starting coordinates of mouse drag.
        """
        event = self.get_mouse_event(event)
        self.lastx, self.lasty = event.xdata, event.ydata
        if event.inaxes:
            self.pressed = event.button
        return None

    def mouseMoveEvent(self, event):
        """
        Translate axes focus while dragging. If there are subplots, only the subplot in which
        cursor exists will be translated.
        """
        if (
            self.pressed not in (MouseButton.LEFT, MouseButton.RIGHT)
            or self.lastx is None
        ):
            return None
        fig = self.figure

        event = self.get_mouse_event(event)

        for ax in fig.axes:
            if event.inaxes != ax:
                continue
            dx = event.xdata - self.lastx
            dy = event.ydata - self.lasty

            x0, x1 = ax.get_xlim()
            y0, y1 = ax.get_ylim()
            if self.pressed is MouseButton.LEFT:
                ax.set_xlim([x0 - dx, x1 - dx])
                ax.set_ylim([y0 - dy, y1 - dy])
            elif self.pressed is MouseButton.RIGHT:
                ax.set_xlim([x0, x1 - dx])
                ax.set_ylim([y0, y1 - dy])
            break

        fig.canvas.draw()
        return None

    def mouseReleaseEvent(self, event):
        """
        Stop dragging state.
        """
        self.pressed = None
        return None

    def mouseDoubleClickEvent(self, event):
        """
        Adjust layout upon dougle click.
        """
        self.figure.tight_layout()
        self.figure.canvas.draw()
        return None

    def resizeEvent(self, event):
        """
        Adjust layout upon canvas resized.
        """
        super().resizeEvent(event)
        self.figure.tight_layout()
        self.figure.canvas.draw()
        return None

    def get_mouse_event(self, event, name="") -> MouseEvent:
        x, y = self.mouseEventCoords(event)
        if hasattr(event, "button"):
            button = self.buttond.get(event.button())
        else:
            button = None
        mouse_event = MouseEvent(name, self, x, y, button=button, guiEvent=event)
        return mouse_event



from .step import (
    GaussStep,
    SDFixedGaussStep,
    TtestStep,
    PoissonStep,
    BayesianPoissonStep,
    BaseStep,
)

__all__ = [
    "GaussStep",
    "SDFixedGaussStep",
    "TtestStep",
    "PoissonStep",
    "BayesianPoissonStep",
    "BaseStep",
]



from __future__ import annotations

from hmmlearn import hmm
import numpy as np

__all__ = ["hmm_sampling", "motor_sampling"]


def hmm_sampling(
    dim: int = 3,
    n_data: int = 500,
    trs: float = 0.05,
    sigma: float = 0.5,
    rand: int | None = None,
    ans: bool = False,
    scale: float = 1,
    poi: bool = False,
):
    """
    Sampline function.py

    Parameters
    ----------
    dim : int, default is 3
        The number of states.
    n_data : int, default is 500
        The length of data.
    trs : float, default is 0.05
        Probability of transition.
    sigma : float, default is 0.5
        Standard deviation of noise.
    rand : int or None, optional
        Random seed.
    ans : bool, default is False
        If the answer of state sequence is returned.
    scale : int, default is 1
        Interval between mean values.
    poi : bool, default is False
        If Poisson distributed.

    """
    startprob = np.full(dim, 1.0 / dim)
    transmat = np.full((dim, dim), trs / (dim - 1)) + np.identity(dim) * (
        1.0 - trs / (dim - 1) - trs
    )
    means = np.arange(1, dim + 1).reshape(-1, 1) * scale
    covars = np.full(dim, sigma * sigma) * scale * scale

    model = hmm.GaussianHMM(n_components=dim, covariance_type="spherical")
    model.startprob_ = startprob
    model.transmat_ = transmat
    model.means_ = means
    model.covars_ = covars

    data, states = model.sample(n_data, random_state=rand)

    answer = model.means_[states, 0]

    if poi:
        np.random.seed(rand)
        data_1 = np.random.poisson(lam=answer)
        np.random.seed()
    else:
        data_1 = data.flatten()

    if ans:
        return data_1, answer
    else:
        return data_1


def motor_sampling(
    pdf=[0.005, -1, 0.015],
    sigma: float = 0.5,
    n_data: int = 500,
    rand: int | None = None,
    ans: bool = False,
):
    """
    pdf: probability distribution
        [... , 2_steps_backward, 1_step_backward, stay, 1_step_forward, 2_steps_forward,
        ...]. one of pdf can be -1
    """
    pdf = np.array(pdf)
    if np.any(pdf == -1):
        rem = 1 - np.sum(pdf[pdf >= 0])
        pdf[pdf == -1] = rem

    if np.sum(pdf) != 1:
        raise ValueError("sum of pdf must be 1")

    np.random.seed(seed=rand)

    rands = np.random.rand(n_data)
    cum_pdf = np.cumsum(pdf).reshape(-1, 1)
    assign_mat, _ = np.meshgrid(rands, np.arange(len(pdf))) > cum_pdf

    delta_y = np.sum(assign_mat.astype("int"), axis=0) - (len(pdf) - 1) / 2
    y_m = np.cumsum(delta_y)
    y = y_m + np.random.normal(scale=sigma, size=n_data)
    np.random.seed()
    if ans:
        return y, y_m
    else:
        return y



from abc import abstractmethod
import os
import numpy as np
from numpy.typing import NDArray
import matplotlib.pyplot as plt
from hmmlearn.hmm import GaussianHMM
from sfHMM.utils import gauss_mix
from sfHMM.gmm import GMMs, DPGMM
from sfHMM.step import GaussStep, PoissonStep, BaseStep


class sfHMMBase(GaussianHMM):
    count = 0
    colors = {
        "raw data": "#FF89F4",
        "step finding": "#335426",
        "denoised": "#180CB4",
        "GMM": "#ED1111",
        "GMM marker": "#FF81D8",
        "Viterbi path": "#3B252B",
    }
    styles = {
        "font.size": 16,
        "lines.linewidth": 1,
        "axes.titlesize": 24,
        "font.family": "serif",
        "font.serif": "Arial",
        "axes.grid": True,
        "axes.labelsize": 16,
        "grid.linewidth": 0.5,
        "legend.framealpha": 0.8,
        "legend.frameon": False,
        "boxplot.meanprops.linewidth": 1,
    }

    means_: NDArray[np.number]
    startprob_: NDArray[np.number]
    covars_: NDArray[np.number]
    transmat_: NDArray[np.number]

    def __init__(
        self,
        sg0: float = -1,
        psf: float = -1,
        krange: "int | tuple[int, int] | None" = None,
        model: str = "g",
        name: str = "",
        **hmmlearn_params,
    ):
        self.sg0 = sg0
        self.psf = psf
        self.krange = krange
        self.model = model
        params = dict(covariance_type="spherical", init_params="")
        params.update(hmmlearn_params)
        super().__init__(self, **params)
        self.n_features = 1
        self.name = str(name) if name else self._name()
        self._log: list[tuple[str, str, str]] = []

    @property
    def log(self):
        out = ""
        for func, n, description in self._log:
            if n == 1:
                out += f"{func}: {description}\n"
            else:
                out += f"{func} ({n}): {description}\n"
        return out

    @property
    def krange(self):
        return self._krange

    @krange.setter
    def krange(self, value):
        if value is None:
            pass
        elif isinstance(value, int):
            value = max(1, value)
            value = (value, value)
        else:
            kmin, kmax = value
            kmin = max(1, kmin)
            kmax = max(kmin, kmax)
            value = (kmin, kmax)
        self._krange = value

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, s):
        if isinstance(s, str):
            if s.lower() in ("g", "gauss", "gaussstep"):
                self._model = "GaussStep"
                self.StepClass = GaussStep
            elif s.lower() in ("p", "poisson", "poissonstep"):
                self._model = "PoissonStep"
                self.StepClass = PoissonStep
            else:
                raise ValueError(f"Invalid model identifier: {s}")

        elif issubclass(s, BaseStep):
            self._model = s.__name__
            self.StepClass = s

        else:
            raise TypeError(
                "`model` must be string or BaseStep instance, " f"but got {type(s)}"
            )

    def get_params(self, deep: bool = True):
        out = super().get_params(deep=deep)
        out.pop("data_raw", None)  # This is not parameter
        out.pop("name", None)  # This does not affect analysis
        return out

    def save(self, path: str = None, overwrite: bool = False):
        """
        Save the content.

        Parameters
        ----------
        path : str, optional
            Saving path. If not given, data will saved at the same directory where it was read.
        overwrite : bool, default is False
            Allow overwriting existing file.

        """
        from sfHMM.io import save

        if path is None:
            try:
                source = self.source
            except AttributeError:
                raise AttributeError(
                    "Data was not read by 'read' function so that the location of "
                    "original data is unknown. 'path' argument is needed."
                )
            file, ext = os.path.splitext(source)
            path = "".join([file, "-sfHMMresult", ext])
        if os.path.exists(path) and not overwrite:
            raise FileExistsError(
                f"File {path} already exists. Change the name or set 'overwrite=True'."
            )
        save(self, path)
        return None

    @abstractmethod
    def step_finding(self): ...

    @abstractmethod
    def denoising(self): ...

    @abstractmethod
    def gmmfit(self, method, n_init, random_state): ...

    @abstractmethod
    def hmmfit(self): ...

    @abstractmethod
    def plot(self): ...

    def run_all(self, plot: bool = True, continue_: bool = False):
        """
        Conduct all the processes with default settings.

        Parameters
        ----------
        plot : bool, default is True
            Specify if show the plot of each process, by default True.
        continue_ : bool, default is False
            If True, and sfHMM analysis is half-way, then only run the rest of analysis.
        """
        if continue_:
            logs = [_l[0] for _l in self._log if _l[2] == "Passed"]
        else:
            logs = []

        for func in ["step_finding", "denoising", "gmmfit", "hmmfit"]:
            if func not in logs:
                getattr(self, func)()
        plot and self.plot()

        return self

    def accumulate_transitions(self):
        """
        This function returns all the transitions occurred in the trajectory.
        """

    def fit(self, X, lengths=None):
        # overloaded to make it scalable.
        scale = self.sg0 * 5
        scale = 1.0 if scale <= 0 else scale

        self.means_ /= scale
        self._covars_ /= scale**2
        super().fit(X / scale, lengths=lengths)
        self.means_ *= scale
        self._covars_ *= scale**2
        return self

    def _name(self):
        self.__class__.count += 1
        return f"{self.__class__.__name__}-{self.__class__.count - 1}"

    def _hist(self, sl=None, ylim=None):
        """
        Draw a histogram that is composed of raw data, denoised data and fitting curve of GMM.
        """
        if sl is None:
            sl = slice(None)
            ylim = self.ylim
            amp = 1
        elif isinstance(sl, slice):
            amp = self.data_raw.size / (sl.stop - sl.start)
        else:
            raise TypeError("sl must be a slice object.")
        plt.ylim(ylim)
        n_bin = min(int(np.sqrt(self.data_raw[sl].size * 2)), 256)
        fit_x = np.linspace(self.ylim[0], self.ylim[1], 256)
        fit_y = gauss_mix(fit_x, self.gmm_opt) * amp
        peak_x = self.gmm_opt.means_.ravel()
        peak_y = gauss_mix(peak_x, self.gmm_opt) * amp
        peak_y += np.max(peak_y) * 0.1
        plt.plot(fit_y, fit_x, color=self.colors["GMM"], linestyle="-.")
        plt.plot(
            peak_y,
            peak_x,
            "<",
            color=self.colors["GMM"],
            markerfacecolor=self.colors["GMM marker"],
            markersize=10,
        )
        plt.hist(
            self.data_raw[sl],
            bins=n_bin,
            color=self.colors["raw data"],
            orientation="horizontal",
            alpha=0.7,
            density=True,
        )
        plt.hist(
            self.data_fil[sl],
            bins=n_bin,
            color=self.colors["denoised"],
            orientation="horizontal",
            histtype="step",
            density=True,
            lw=2,
        )

        return None

    def _init_sg0(self, p: float = 25):
        """
        Initialize 'sg0' if it is negative.
        """
        if self.sg0 < 0:
            _l = np.abs(self._accumulate_step_sizes())
            if len(_l) > 0:
                self.sg0 = np.percentile(_l, p) * 0.2
            else:
                self.sg0 = np.std(self.data_raw)

        return None

    def _gmmfit(self, method: str, n_init: int, random_state: int):
        # in case S.D. of noise was very small
        if len(self._sg_list) > 0:
            sg0_ = min(self.sg0, np.percentile(self._sg_list, 25))
        else:
            sg0_ = self.sg0

        if self.krange is None:
            self.krange = (1, 6)  # Enough for most situations.

        if method.lower() in ("aic", "bic"):
            gmm_ = GMMs(
                self.data_fil,
                self.krange,
                min_interval=sg0_ * 1.5,
                min_sg=sg0_ * 0.8,
                covariance_type=self.covariance_type,
            )
            gmm_.fit(n_init=n_init, random_state=random_state, scale=self.sg0 * 5)
            self.gmm = gmm_
            self.gmm_opt = self.gmm.get_optimal(method)
        elif method.lower() == "dirichlet":
            gmm_ = DPGMM(
                n_components=self.krange[1],
                n_init=1,
                random_state=random_state,
                covariance_prior=0.02,
                covariance_type=self.covariance_type,
            )
            gmm_.fit(np.asarray(self.data_fil).reshape(-1, 1), scale=self.sg0 * 5)
            self.gmm_opt = gmm_
        else:
            raise ValueError(f"method: {method}")

        self.n_components = self.gmm_opt.n_components

        return None

    def tdp(self, **kwargs):
        """
        Pseudo transition density plot.
        """
        means = self.means_.ravel()
        cov = self.covars_.ravel()

        axlim = (np.min(means) - np.sqrt(cov.max()), np.max(means) + np.sqrt(cov.max()))

        tr = self.accumulate_transitions()

        axes = np.linspace(*axlim, 200)
        x, y = np.meshgrid(axes, axes)
        z = np.zeros((200, 200))

        for sx, sy in tr:
            mx = self.means_[sx]
            my = self.means_[sy]
            z += np.exp(
                -((x - mx) ** 2 / (2 * cov[sx]) + (y - my) ** 2 / (2 * cov[sy]))
            )

        z /= np.max(z)

        kw = {"vmin": 0, "cmap": "jet", "origin": "lower"}
        kw.update(kwargs)

        with plt.style.context(self.__class__.styles):
            plt.figure()
            plt.title("Transition Density Plot")
            plt.imshow(z.T, **kw)
            plt.colorbar()
            pos = ((means - axlim[0]) / (axlim[1] - axlim[0]) * 200).astype("int16")
            digit_0 = int(np.median(np.floor(np.log10(np.abs(means)))))
            plt.xticks(pos, np.round(means, -digit_0 + 1))
            plt.yticks(pos, np.round(means, -digit_0 + 1))
            plt.xlabel("Before")
            plt.ylabel("After")
            plt.show()
        return None

    def _set_hmm_params(self):
        if not hasattr(self, "n_components"):
            raise AttributeError("'n_components' has yet been specified.")

        hasattr(self, "covars_") or self._set_covars()
        hasattr(self, "means_") or self._set_means()
        hasattr(self, "startprob_") or self._set_startprob()
        hasattr(self, "transmat_") or self._set_transmat()

        return None



from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from warnings import warn
from typing import TypeVar
from .base import sfHMMBase
from .utils import (
    append_log,
    plot2,
    calc_covars,
    calc_startprob,
    calc_transmat,
    sfHMMAnalysisError,
)

__all__ = ["sfHMM1"]

_S = TypeVar("_S")  # array-like


class sfHMM1(sfHMMBase):
    """
    Step-finding based HMM for single trajectory.

    Analysis Results
    ----------------
    data_raw : np.ndarray
        Raw data.
    step : `GaussStep` or `PoissonStep` object
        The result of step finding, which has following attributes:
        - fit ... Fitting result.
        - n_step ... The number of steps (region between two signal change points).
        - step_list ... List of signal change points.
        - mu_list ... list of mean values of each step.
        - len_list ... list of step lengths (step_list[i+1] - step_list[i]).
        - step_size_list ... list of signal change (mu_list[i+1] - mu_list[i]).
    data_fil : np.ndarray
        Data after denoised.
    gmm_opt : `GMMs` or `DPGMM` object. For more detail, see `sfHMM.gmm`.
        The result of GMM clustering, which inherits `sklearn.mixture.GaussianMixture`.
        If AIC/BIC minimization of standard GMM clustering was conducted, the clustering
        results will be stored in `gmm`. See `sfHMM.gmm.GMMs`.
    n_components : int
        The optimal number of states. Same as 'gmm_opt.n_components'.
    states : np.ndarray
        The optimal state sequence. Before HMM fitting, this is determined from the
        results of step finding and GMM clustering. After HMM fitting, this is Viterbi
        path with values {0, 1, 2, ...}.
    viterbi : np.ndarray
        Viterbi path of 'data_raw', while takes values in 'means_'.
    """

    def __init__(
        self,
        data_raw: _S | None = None,
        *,
        sg0: float = -1,
        psf: float = -1,
        krange: int | tuple[int, int] | None = None,
        model: str = "g",
        name: str = "",
        **kwargs,
    ):
        """
        Parameters
        ----------
        data_raw : array like, optional
            Data for analysis.
        sg0 : float, optional
            Parameter used in filtering method. Expected to be 20% of signal change.
            If <= 0, sg0 will be determined automatically.
        psf : float, optional
            Transition probability used in step finding algorithm.
            if 0 < p < 0.5 is not satisfied, the original Kalafut-Visscher's algorithm
            is executed.
        krange : int or (int, int)
            Minimum and maximum number of states to search in GMM clustering. If it is
            integer, then it will be interpretted as (krange, krange).
        model: str, by default "g" (= Gaussian)
            Distribution of noise. Gauss and Poisson distribution are available for now.
        name : str, optional
            Name of the object.
        """
        self.step = None
        self.data_fil = None
        self.gmm_opt = None
        self.states = None
        self.viterbi = None
        self._sg_list: list[float] = []
        super().__init__(sg0, psf, krange, model, name, **kwargs)
        self.data_raw = data_raw

    @property
    def size(self) -> int:
        return self.data_raw.size

    @property
    def data_raw(self) -> np.ndarray:
        return self._data_raw

    @data_raw.setter
    def data_raw(self, value: _S):
        if value is None:
            self._data_raw = None
        elif np.isscalar(value):
            raise TypeError(f"Wrong type of input data: {type(value)}")
        else:
            d = np.asarray(value)
            if not np.issubdtype(d.dtype, np.number):
                raise TypeError("Input contains non-numeric object(s).")
            elif d.ndim == 1:
                pass
            elif d.ndim == 2 and (d.shape[0] == 1 or d.shape[1] == 1):
                d = d.ravel()
            elif self.krange is not None and d.size < self.krange[1]:
                raise ValueError(f"Input data size is too small: {d.size}")
            else:
                raise ValueError(
                    "Input data must be one-dimensonal or any arrays "
                    "that can be converted to one-dimensional ones."
                )
            self._data_raw = d
            self.ylim = [np.min(self.data_raw), np.max(self.data_raw)]

    def read(
        self,
        path: str,
        sep: str = None,
        encoding: str = None,
        header: int = 0,
        **kwargs,
    ):
        """
        Read a file using appropriate function, and import its data to sfHMM1 object. If
        multiple trajectories are found, ValueError will be raised.

        Parameters
        ----------
        path : str
            Path to file.
        sep, encoding, header
            Important arguments in pd.read_csv. Default is header=0 rather than "infer"
            because header="infer" usually works in a wrong way.
        **kwargs
            Other keyword arguments that will passed to pd.read_csv or pd.read_excel.
        """
        from .io import read

        out = read(path, sep=sep, encoding=encoding, header=header, **kwargs)
        if len(out._sf_list) > 1:
            raise ValueError("More than one trajectory found. Use sfHMMn instead.")
        self.data_raw = out[0].data_raw
        self.source = path
        return self

    @append_log
    def step_finding(self) -> sfHMM1:
        """
        Step finding by extended version of Kalafut-Visscher's algorithm.
        """
        if np.all(np.diff(self.data_raw) > 0):
            warn(
                f"Data of {self.name} is monotonically increasing. "
                "Isn't it a time axis?",
                UserWarning,
                stacklevel=2,
            )
        self.step = self.StepClass(self.data_raw, self.psf)
        self.step.multi_step_finding()
        self.psf = getattr(self.step, "p", -1)
        return self

    @append_log
    def denoising(self) -> sfHMM1:
        """
        Denoising by cutting of the standard deviation of noise to sg0.
        """
        if self.step is None:
            raise sfHMMAnalysisError("Cannot run denoising before step finding.")

        self._init_sg0()
        self.data_fil = np.empty_like(self.data_raw, dtype=np.float64)

        for i in range(self.step.n_step):
            x0 = self.step.step_list[i]
            x1 = self.step.step_list[i + 1]
            mu = self.step.mu_list[i]
            sg = np.sqrt(np.mean((self.data_raw[x0:x1] - mu) ** 2))
            self._sg_list.append(sg)
            if self.sg0 < sg:
                self.data_fil[x0:x1] = (self.data_raw[x0:x1] - mu) * self.sg0 / sg + mu
            else:
                self.data_fil[x0:x1] = self.data_raw[x0:x1]

        return self

    @append_log
    def gmmfit(
        self, method: str = "bic", n_init: int = 1, random_state: int = 0
    ) -> sfHMM1:
        """
        Fit the denoised data to Gaussian mixture model, and the optimal number of
        states will be determined. After that, state sequence 'states' will be
        initialized.

        Parameters
        ----------
        method : str, optional
            How to determine the optimal number of states. This parameter must be
            'aic', 'bic' or 'Dirichlet'. By default "bic".
        n_init : int, optional
            How many times initialization will be performed in K-means, by default 1.
        random_state : int , optional
            Random seed for K-means initialization., by default 0.

        """
        # If denoising was passed.
        if self.data_fil is None:
            self.data_fil = self.data_raw

        # Start GMM clustering and determine optimal number of states.
        self._gmmfit(method, n_init, random_state)

        # If denoising is conducted without step finding, state sequence will be
        # inferred using 'self.data_fil'.
        if self.step is not None:
            self.states = self.gmm_opt.predict(np.asarray(self.step.fit).reshape(-1, 1))
        else:
            self.states = self.gmm_opt.predict(np.asarray(self.data_fil).reshape(-1, 1))

        return self

    @append_log
    def hmmfit(self) -> sfHMM1:
        """
        HMM paramter optimization by EM algorithm, and state inference by Viterbi
        algorithm.
        """
        self._set_hmm_params()

        _data_reshaped = np.asarray(self.data_raw).reshape(-1, 1)
        self.fit(_data_reshaped)
        self.states = self.predict(_data_reshaped)
        self.viterbi = self.means_[self.states, 0]

        return self

    def plot(self, trange: int | tuple[int, int] | None = None):
        """
        Plot figures of:
            [1] data_raw & step_fit      ||  layout
            [2] data_raw & data_fil      ||  [ 1 ]
            [3] histograms of [2]        ||  [ 2 ][3]
            [4] data_raw & viterbi       ||  [ 4 ]

        Parameters
        ----------
        trange : array like, optional
            Range of x-axis to show, by default None
        """

        if trange is None:
            sl = slice(0, self.size)
            ylim = self.ylim
        elif isinstance(trange, (list, tuple, np.ndarray)):
            sl = slice(*trange)
            ylim = np.min(self.data_raw[sl]), np.max(self.data_raw[sl])
        elif isinstance(trange, slice):
            sl = trange
            ylim = np.min(self.data_raw[sl]), np.max(self.data_raw[sl])
        else:
            raise TypeError(
                "`trange` must be a slice or an array-like object, but got "
                f"{type(trange)}."
            )

        tasks = []
        showhist = self.gmm_opt is not None
        self.step is None or tasks.append("step finding")
        self.data_fil is None or tasks.append("denoised")
        self.viterbi is None or tasks.append("Viterbi path")
        c_raw = self.__class__.colors["raw data"]
        n_row = max(len(tasks), 1)
        n_col = showhist + 1

        with plt.style.context(self.__class__.styles):
            plt.figure(figsize=(6 * n_col, 4.2 * n_row))

            for i, task in enumerate(tasks):
                i += 1
                plt.subplot(n_row, n_col, (i - 1) * n_col + 1)
                i == 1 and plt.title(self.name, fontweight="bold")
                kw = dict(
                    ylim=ylim,
                    color1=c_raw,
                    color=self.__class__.colors[task],
                    label=task,
                )
                if task == "step finding":
                    plot2(self.data_raw[sl], self.step.fit[sl], **kw)
                elif task == "denoised":
                    plot2(self.data_raw[sl], self.data_fil[sl], legend=False, **kw)
                    if showhist:
                        plt.subplot(n_row, n_col * 2, 2 * (i - 1) * n_col + 3)
                        self._hist(sl, ylim)
                elif task == "Viterbi path":
                    plot2(self.data_raw[sl], self.viterbi[sl], **kw)
                else:
                    raise NotImplementedError

            if trange is not None:
                ax = plt.subplot(6, 4, 7)
                ax.plot(self.data_raw, color="gray", alpha=0.3)
                ax.plot(
                    np.arange(sl.start, sl.stop),
                    self.data_raw[sl],
                    color=self.__class__.colors["raw data"],
                )
                ax.set_xlim(0, self.size)
                ax.set_ylim(self.ylim)
                ax.plot(
                    [sl.start, sl.stop, sl.stop, sl.start, sl.start],
                    [ylim[0], ylim[0], ylim[1], ylim[1], ylim[0]],
                    color="black",
                )
                ax.set_xticks([])
                ax.set_yticks([])

            len(tasks) == 0 and plot2(self.data_raw[sl], ylim=ylim, color1=c_raw)

            plt.show()

        return None

    def view_in_qt(self, title: str | None = None):
        """
        Open a Qt viewer and plot the results.
        """
        from .viewer import TrajectoryViewer

        if title is None:
            title = self.name
        data = dict()
        self.data_raw is None or data.update({"raw data": self.data_raw})
        self.step is None or data.update({"step finding": self.step.fit})
        self.data_fil is None or data.update({"denoised": self.data_fil})
        self.viterbi is None or data.update({"Viterbi path": self.viterbi})
        viewer = TrajectoryViewer(data, self.__class__.styles, self.__class__.colors)
        viewer.setWindowTitle(title)
        viewer.show()
        return viewer

    def accumulate_transitions(self) -> list[tuple[int, int]]:
        """
        Accumulate all the transitions occurred in `self.states`, and return them in
        [(y0, y1), (y1, y2), (y2, y3), ...] format.

        Returns
        -------
        list[tuple[int, int]]
            List of transitions.
        """
        if not hasattr(self, "states"):
            return np.array([], dtype=np.float64)
        return [
            (self.states[i], self.states[i + 1])
            for i in range(self.states.size - 1)
            if self.states[i] != self.states[i + 1]
        ]

    def _accumulate_step_sizes(self):
        if self.step is None:
            raise sfHMMAnalysisError("Steps are not detected yet.")
        return self.step.step_size_list

    def _set_covars(self):
        if self.states is None:
            raise sfHMMAnalysisError(
                "Cannot initialize 'covars_' because the state sequence "
                "'states' hasyet been determined."
            )
        self.covars_ = calc_covars(self.data_raw, self.states, self.n_components)
        return None

    def _set_means(self):
        if self.gmm_opt is None:
            raise sfHMMAnalysisError(
                "Cannot initialize 'means_'. You must run gmmfit() before "
                "hmmfit() or set 'means_' manually."
            )
        self.means_ = self.gmm_opt.means_.copy()
        return None

    def _set_startprob(self):
        if self.gmm_opt is None:
            raise sfHMMAnalysisError(
                "Cannot initialize 'startprob_'. You must run gmmfit() "
                "before hmmfit() or set 'startprob_' manually."
            )
        self.startprob_ = calc_startprob([self.data_raw[0]], self.gmm_opt)
        return None

    def _set_transmat(self):
        if self.states is None:
            raise sfHMMAnalysisError(
                "Cannot initialize 'transmat_' because the state sequence "
                "'states' has yet been determined."
            )
        self.transmat_ = calc_transmat([self.states], self.n_components)
        return None
